"""
Test the definition of attributes and their constraints.
"""

from __future__ import annotations
from dataclasses import dataclass
from io import StringIO
from typing import Any, TypeVar, cast, Annotated, Generic, TypeAlias

import pytest

from xdsl.ir import Attribute, Data, ParametrizedAttribute
from xdsl.irdl import (AttrConstraint, GenericData, ParameterDef,
                       irdl_attr_definition, builder, irdl_to_attr_constraint,
                       AnyAttr, BaseAttr, ParamAttrDef)
from xdsl.parser import BaseParser
from xdsl.printer import Printer
from xdsl.utils.exceptions import VerifyException

#  ____        _
# |  _ \  __ _| |_ __ _
# | | | |/ _` | __/ _` |
# | |_| | (_| | || (_| |
# |____/ \__,_|\__\__,_|
#


@irdl_attr_definition
class BoolData(Data[bool]):
    """An attribute holding a boolean value."""
    name = "bool"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> bool:
        val = parser.tokenizer.next_token_of_pattern('(True|False)')
        if val is None or val.text not in ('True', 'False'):
            parser.raise_error("Expected True or False literal")
        if val.text == "True":
            return True
        return False

    @staticmethod
    def print_parameter(data: bool, printer: Printer):
        printer.print_string(str(data))


@irdl_attr_definition
class IntData(Data[int]):
    """An attribute holding an integer value."""
    name = "int"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> int:
        return parser.parse_int_literal()

    @staticmethod
    def print_parameter(data: int, printer: Printer):
        printer.print_string(str(data))


@irdl_attr_definition
class StringData(Data[str]):
    """An attribute holding a string value."""
    name = "str"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> str:
        return parser.parse_str_literal()

    @staticmethod
    def print_parameter(data: str, printer: Printer):
        printer.print_string(data)


def test_simple_data():
    """Test that the definition of a data with a class parameter."""
    b = BoolData(True)
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(b)
    assert stream.getvalue() == "!bool<True>"


def test_simple_data_verifier_failure():
    """
    Test that the verifier of a data with a class parameter fails when given
    a parameter of the wrong type.
    """
    with pytest.raises(VerifyException) as e:
        BoolData(2)  # type: ignore
    assert e.value.args[0] == ("bool data attribute expected type "
                               "<class 'bool'>, but <class 'int'> given.")


class IntListMissingVerifierData(Data[list[int]]):
    """
    An attribute holding a list of integers.
    The definition should fail, since no verifier is provided, and the Data
    type parameter is not a class.
    """
    name = "missing_verifier_data"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> list[int]:
        raise NotImplementedError()

    @staticmethod
    def print_parameter(data: list[int], printer: Printer) -> None:
        raise NotImplementedError()


def test_data_with_non_class_param_missing_verifier_failure():
    """
    Test that a non-class Data parameter requires the definition of a verifier.
    """
    with pytest.raises(Exception) as e:
        irdl_attr_definition(IntListMissingVerifierData)

    # Python 3.10 and 3.11 have different error messages
    assert e.value.args[0] in [
        'In IntListMissingVerifierData definition: '
        'Cannot infer "verify" method. Type parameter of Data has type GenericAlias.',
        'In IntListMissingVerifierData definition: '
        'Cannot infer "verify" method. Type parameter of Data is not a class.',
    ]


@irdl_attr_definition
class IntListData(Data[list[int]]):
    """
    An attribute holding a list of integers.
    """
    name = "int_list"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> list[int]:
        raise NotImplementedError()

    @staticmethod
    def print_parameter(data: list[int], printer: Printer) -> None:
        printer.print_string("[")
        printer.print_list(data, lambda x: printer.print_string(str(x)))
        printer.print_string("]")

    def verify(self) -> None:
        if not isinstance(self.data, list):
            raise VerifyException("int_list data should hold a list.")
        for elem in self.data:
            if not isinstance(elem, int):
                raise VerifyException(
                    "int_list list elements should be integers.")


def test_non_class_data():
    """Test the definition of a Data with a non-class parameter."""
    attr = IntListData([0, 1, 42])
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "!int_list<[0, 1, 42]>"


def test_simple_data_constructor_failure():
    """
    Test that the verifier of a Data with a non-class parameter fails when
    given wrong arguments.
    """
    with pytest.raises(VerifyException) as e:
        IntListData([0, 1, 42, ""])  # type: ignore
    assert e.value.args[0] == "int_list list elements should be integers."


#  ____                  ____                _             _       _
# | __ )  __ _ ___  ___ / ___|___  _ __  ___| |_ _ __ __ _(_)_ __ | |_
# |  _ \ / _` / __|/ _ \ |   / _ \| '_ \/ __| __| '__/ _` | | '_ \| __|
# | |_) | (_| \__ \  __/ |__| (_) | | | \__ \ |_| | | (_| | | | | | |_
# |____/ \__,_|___/\___|\____\___/|_| |_|___/\__|_|  \__,_|_|_| |_|\__|
#


@irdl_attr_definition
class BoolWrapperAttr(ParametrizedAttribute):
    name = "bool_wrapper"

    param: ParameterDef[BoolData]


def test_bose_constraint():
    """Test the verifier of a base attribute type constraint."""
    attr = BoolWrapperAttr([BoolData(True)])
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "!bool_wrapper<!bool<True>>"


def test_base_constraint_fail():
    """Test the verifier of a union constraint."""
    with pytest.raises(Exception) as e:
        BoolWrapperAttr([StringData("foo")])
    assert e.value.args[0] == "!str<foo> should be of base attribute bool"


#  _   _       _              ____                _             _       _
# | | | |_ __ (_) ___  _ __  / ___|___  _ __  ___| |_ _ __ __ _(_)_ __ | |_
# | | | | '_ \| |/ _ \| '_ \| |   / _ \| '_ \/ __| __| '__/ _` | | '_ \| __|
# | |_| | | | | | (_) | | | | |__| (_) | | | \__ \ |_| | | (_| | | | | | |_
#  \___/|_| |_|_|\___/|_| |_|\____\___/|_| |_|___/\__|_|  \__,_|_|_| |_|\__|
#


@irdl_attr_definition
class BoolOrIntParamAttr(ParametrizedAttribute):
    name = "bool_or_int"

    param: ParameterDef[BoolData | IntData]


def test_union_constraint_left():
    """Test the verifier of a union constraint."""
    attr = BoolOrIntParamAttr([BoolData(True)])
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "!bool_or_int<!bool<True>>"


def test_union_constraint_right():
    """Test the verifier of a union constraint."""
    attr = BoolOrIntParamAttr([IntData(42)])
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "!bool_or_int<!int<42>>"


def test_union_constraint_fail():
    """Test the verifier of a union constraint."""
    with pytest.raises(Exception) as e:
        BoolOrIntParamAttr([StringData("foo")])
    assert e.value.args[0] == "Unexpected attribute !str<foo>"


#     _                      _    ____                _
#    / \   _ __  _ __   ___ | |_ / ___|___  _ __  ___| |_ _ __
#   / _ \ | '_ \| '_ \ / _ \| __| |   / _ \| '_ \/ __| __| '__|
#  / ___ \| | | | | | | (_) | |_| |__| (_) | | | \__ \ |_| |
# /_/   \_\_| |_|_| |_|\___/ \__|\____\___/|_| |_|___/\__|_|


class PositiveIntConstr(AttrConstraint):

    def verify(self, attr: Attribute) -> None:
        if not isinstance(attr, IntData):
            raise VerifyException(
                f"Expected {IntData.name} attribute, but got {attr.name}.")
        if attr.data <= 0:
            raise VerifyException(
                f"Expected positive integer, got {attr.data}.")


@irdl_attr_definition
class PositiveIntAttr(ParametrizedAttribute):
    name = "positive_int"

    param: ParameterDef[Annotated[IntData, PositiveIntConstr()]]


def test_annotated_constraint():
    """Test the verifier of an annotated constraint."""
    attr = PositiveIntAttr([IntData(42)])
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "!positive_int<!int<42>>"


def test_annotated_constraint_fail():
    """Test that the verifier of an annotated constraint can fail."""
    with pytest.raises(Exception) as e:
        PositiveIntAttr([IntData(-42)])
    assert e.value.args[0] == "Expected positive integer, got -42."


#  _____               __     __          ____                _
# |_   _|   _ _ __   __\ \   / /_ _ _ __ / ___|___  _ __  ___| |_ _ __
#   | || | | | '_ \ / _ \ \ / / _` | '__| |   / _ \| '_ \/ __| __| '__|
#   | || |_| | |_) |  __/\ V / (_| | |  | |__| (_) | | | \__ \ |_| |
#   |_| \__, | .__/ \___| \_/ \__,_|_|   \____\___/|_| |_|___/\__|_|
#       |___/|_|
#

_T = TypeVar("_T", bound=BoolData | IntData)


@irdl_attr_definition
class ParamWrapperAttr(Generic[_T], ParametrizedAttribute):
    name = "int_or_bool_generic"

    param: ParameterDef[_T]


def test_typevar_attribute_int():
    """Test the verifier of a generic attribute."""
    attr = ParamWrapperAttr[IntData]([IntData(42)])
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "!int_or_bool_generic<!int<42>>"


def test_typevar_attribute_bool():
    """Test the verifier of a generic attribute."""
    attr = ParamWrapperAttr[BoolData]([BoolData(True)])
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "!int_or_bool_generic<!bool<True>>"


def test_typevar_attribute_fail():
    """Test that the verifier of an generic attribute can fail."""
    with pytest.raises(Exception) as e:
        ParamWrapperAttr([StringData("foo")])
    assert e.value.args[0] == "Unexpected attribute !str<foo>"


@irdl_attr_definition
class ParamConstrAttr(ParametrizedAttribute):
    name = "param_constr"

    param: ParameterDef[ParamWrapperAttr[IntData]]


def test_param_attr_constraint():
    """Test the verifier of an attribute with a parametric constraint."""
    attr = ParamConstrAttr([ParamWrapperAttr([IntData(42)])])
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "!param_constr<!int_or_bool_generic<!int<42>>>"


def test_param_attr_constraint_fail():
    """
    Test that the verifier of an attribute with
    a parametric constraint can fail.
    """
    with pytest.raises(Exception) as e:
        ParamConstrAttr([ParamWrapperAttr([BoolData(True)])])
    assert e.value.args[0] == "!bool<True> should be of base attribute int"


_U = TypeVar("_U", bound=IntData)


@irdl_attr_definition
class NestedParamWrapperAttr(Generic[_U], ParametrizedAttribute):
    name = "nested_param_wrapper"

    param: ParameterDef[ParamWrapperAttr[_U]]


def test_nested_generic_constraint():
    """
    Test the verifier of an attribute with a generic
    constraint used in a parametric constraint.
    """
    attr = NestedParamWrapperAttr[IntData]([ParamWrapperAttr([IntData(42)])])
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue(
    ) == "!nested_param_wrapper<!int_or_bool_generic<!int<42>>>"


def test_nested_generic_constraint_fail():
    """
    Test that the verifier of an attribute with
    a parametric constraint can fail.
    """
    with pytest.raises(Exception) as e:
        NestedParamWrapperAttr([ParamWrapperAttr([BoolData(True)])])
    assert e.value.args[0] == "!bool<True> should be of base attribute int"


@irdl_attr_definition
class NestedParamConstrAttr(ParametrizedAttribute):
    name = "nested_param_constr"

    param: ParameterDef[NestedParamWrapperAttr[Annotated[IntData,
                                                         PositiveIntConstr()]]]


def test_nested_param_attr_constraint():
    """
    Test the verifier of a nested parametric constraint.
    """
    attr = NestedParamConstrAttr(
        [NestedParamWrapperAttr([ParamWrapperAttr([IntData(42)])])])
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue(
    ) == "!nested_param_constr<!nested_param_wrapper<!int_or_bool_generic<!int<42>>>>"


def test_nested_param_attr_constraint_fail():
    """
    Test that the verifier of a nested parametric constraint can fail.
    """
    with pytest.raises(Exception) as e:
        NestedParamConstrAttr(
            [NestedParamWrapperAttr([ParamWrapperAttr([IntData(-42)])])])
    assert e.value.args[0] == "Expected positive integer, got -42."


#   ____                      _      ____        _
#  / ___| ___ _ __   ___ _ __(_) ___|  _ \  __ _| |_ __ _
# | |  _ / _ \ '_ \ / _ \ '__| |/ __| | | |/ _` | __/ _` |
# | |_| |  __/ | | |  __/ |  | | (__| |_| | (_| | || (_| |
#  \____|\___|_| |_|\___|_|  |_|\___|____/ \__,_|\__\__,_|
#

_MissingGenericDataData = TypeVar("_MissingGenericDataData")


@irdl_attr_definition
class MissingGenericDataData(Data[_MissingGenericDataData]):
    name = "missing_genericdata"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> _MissingGenericDataData:
        raise NotImplementedError()

    @staticmethod
    def print_parameter(data: _MissingGenericDataData,
                        printer: Printer) -> None:
        raise NotImplementedError()

    def verify(self) -> None:
        return


class MissingGenericDataDataWrapper(ParametrizedAttribute):
    name = "missing_genericdata_wrapper"

    param: ParameterDef[MissingGenericDataData[int]]


def test_data_with_generic_missing_generic_data_failure():
    """
    Test error message when a generic data is used in constraints
    without implementing GenericData.
    """
    with pytest.raises(Exception) as e:
        irdl_attr_definition(MissingGenericDataDataWrapper)
    assert e.value.args[0] == (
        "Generic `Data` type 'missing_genericdata' cannot be converted to "
        "an attribute constraint. Consider making it inherit from "
        "`GenericData` instead of `Data`.")


A = TypeVar("A", bound=Attribute)


@dataclass
class DataListAttr(AttrConstraint):
    """
    A constraint that enforces that the elements of a ListData all respect
    a constraint.
    """
    elem_constr: AttrConstraint

    def verify(self, attr: Attribute) -> None:
        attr = cast(ListData[Any], attr)
        for e in attr.data:
            self.elem_constr.verify(e)


@irdl_attr_definition
class ListData(GenericData[list[A]]):
    name = "list"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> list[A]:
        raise NotImplementedError()

    @staticmethod
    def print_parameter(data: list[A], printer: Printer) -> None:
        printer.print_string("[")
        printer.print_list(data, printer.print_attribute)
        printer.print_string("]")

    @staticmethod
    def generic_constraint_coercion(args: tuple[Any]) -> AttrConstraint:
        assert len(args) == 1
        return DataListAttr(irdl_to_attr_constraint(args[0]))

    @staticmethod
    @builder
    def from_list(data: list[A]) -> ListData[A]:
        return ListData(data)

    def verify(self) -> None:
        if not isinstance(self.data, list):
            raise VerifyException(
                f"Wrong type given to attribute {self.name}: got"
                f" {type(self.data)}, but expected list of"
                " attributes.")
        for idx, val in enumerate(self.data):
            if not isinstance(val, Attribute):
                raise VerifyException(
                    f"{self.name} data expects attribute list, but element "
                    f"{idx} is of type {type(val)}.")


AnyListData: TypeAlias = ListData[Attribute]


class Test_generic_data_verifier:

    def test_generic_data_verifier(self):
        """
        Test that a GenericData can be created.
        """
        attr = ListData([BoolData(True), ListData([BoolData(False)])])
        stream = StringIO()
        p = Printer(stream=stream)
        p.print_attribute(attr)
        assert stream.getvalue(
        ) == "!list<[!bool<True>, !list<[!bool<False>]>]>"

    def test_generic_data_verifier_fail(self):
        """
        Test that a GenericData verifier fails when given wrong parameters.
        """
        with pytest.raises(VerifyException) as e:
            ListData([0])  # type: ignore
        assert e.value.args[0] == ("list data expects attribute list, but"
                                   " element 0 is of type <class 'int'>.")

    def test_generic_data_verifier_fail_II(self):
        """
        Test that a GenericData verifier fails when given wrong parameters.
        """
        with pytest.raises(VerifyException) as e:
            ListData((0))  # type: ignore
        assert e.value.args[0] == (
            "Wrong type given to attribute list: "
            "got <class 'int'>, but expected list of attributes.")


@irdl_attr_definition
class ListDataWrapper(ParametrizedAttribute):
    name = "list_wrapper"

    val: ParameterDef[ListData[BoolData]]


def test_generic_data_wrapper_verifier():
    """
    Test that a GenericData used in constraints pass the verifier when correct.
    """
    attr = ListDataWrapper([ListData([BoolData(True), BoolData(False)])])
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue(
    ) == "!list_wrapper<!list<[!bool<True>, !bool<False>]>>"


def test_generic_data_wrapper_verifier_failure():
    """
    Test that a GenericData used in constraints fails
    the verifier when constraints are not satisfied.
    """
    with pytest.raises(VerifyException) as e:
        ListDataWrapper(
            [ListData([BoolData(True),
                       ListData([BoolData(False)])])])
    assert e.value.args[
        0] == "!list<[!bool<False>]> should be of base attribute bool"


@irdl_attr_definition
class ListDataNoGenericsWrapper(ParametrizedAttribute):
    name = "list_no_generics_wrapper"

    val: ParameterDef[AnyListData]


def test_generic_data_no_generics_wrapper_verifier():
    """
    Test that GenericType can be used in constraints without a parameter.
    """
    attr = ListDataNoGenericsWrapper(
        [ListData([BoolData(True), ListData([BoolData(False)])])])
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue(
    ) == "!list_no_generics_wrapper<!list<[!bool<True>, !list<[!bool<False>]>]>>"


#  ____                              _   _   _        ____        __
# |  _ \ __ _ _ __ __ _ _ __ ___    / \ | |_| |_ _ __|  _ \  ___ / _|
# | |_) / _` | '__/ _` | '_ ` _ \  / _ \| __| __| '__| | | |/ _ \ |_
# |  __/ (_| | | | (_| | | | | | |/ ___ \ |_| |_| |  | |_| |  __/  _|
# |_|   \__,_|_|  \__,_|_| |_| |_/_/   \_\__|\__|_|  |____/ \___|_|
#


@irdl_attr_definition
class ParamAttrDefAttr(ParametrizedAttribute):
    name = "test.param_attr_def_attr"

    arg1: ParameterDef[Attribute]
    arg2: ParameterDef[BoolData]


def test_irdl_definition():
    """Test that we can get the IRDL definition of a parametrized attribute."""

    assert ParamAttrDefAttr.irdl_definition == ParamAttrDef(
        "test.param_attr_def_attr", [("arg1", AnyAttr()),
                                     ("arg2", BaseAttr(BoolData))])



import pytest

from xdsl.ir import MLContext, Operation, Block, Region
from xdsl.dialects.arith import Addi, Subi, Constant
from xdsl.dialects.builtin import i32, IntegerAttr, ModuleOp
from xdsl.dialects.scf import If
from xdsl.parser import XDSLParser
from xdsl.dialects.builtin import Builtin
from xdsl.dialects.func import Func
from xdsl.dialects.arith import Arith
from xdsl.dialects.cf import Cf


def test_ops_accessor():
    a = Constant.from_int_and_width(1, i32)
    b = Constant.from_int_and_width(2, i32)
    # Operation to add these constants
    c = Addi.get(a, b)

    block0 = Block.from_ops([a, b, c])
    # Create a region to include a, b, c
    region = Region.from_block_list([block0])

    assert len(region.ops) == 3
    assert len(region.blocks[0].ops) == 3

    # Operation to subtract b from a
    d = Subi.get(a, b)

    assert d.results[0] != c.results[0]


def test_ops_accessor_II():
    a = Constant.from_int_and_width(1, i32)
    b = Constant.from_int_and_width(2, i32)
    # Operation to add these constants
    c = Addi.get(a, b)

    block0 = Block.from_ops([a, b, c])
    # Create a region to include a, b, c
    region = Region.from_block_list([block0])

    assert len(region.ops) == 3
    assert len(region.blocks[0].ops) == 3

    # Operation to subtract b from a
    d = Subi.get(a, b)

    assert d.results[0] != c.results[0]

    # Erase operations and block
    region2 = Region()
    region.move_blocks(region2)

    region2.blocks[0].erase_op(a, safe_erase=False)
    region2.blocks[0].erase_op(b, safe_erase=False)
    region2.blocks[0].erase_op(c, safe_erase=False)

    region2.detach_block(block0)
    region2.drop_all_references()

    assert len(region2.blocks) == 0


def test_ops_accessor_III():
    # Create constants `from_attr` and add them, add them in blocks, blocks in
    # a region and create a function
    a = Constant.from_attr(IntegerAttr.from_int_and_width(1, 32), i32)
    b = Constant.from_attr(IntegerAttr.from_int_and_width(2, 32), i32)
    c = Constant.from_attr(IntegerAttr.from_int_and_width(3, 32), i32)
    d = Constant.from_attr(IntegerAttr.from_int_and_width(4, 32), i32)

    # Operation to add these constants
    e = Addi.get(a, b)
    f = Addi.get(c, d)

    # Create Blocks and Regions
    block0 = Block.from_ops([a, b, e])
    block1 = Block.from_ops([c, d, f])
    block2 = Block.from_ops([])

    region0 = Region.from_block_list([block0, block1])
    region1 = Region.from_block_list([block2])

    with pytest.raises(ValueError):
        region0.ops

    with pytest.raises(ValueError):
        region0.op

    with pytest.raises(ValueError):
        region1.op

    with pytest.raises(Exception):
        region1.detach_block(block0)

    region0.detach_block(block0)
    region0.detach_block(0)
    with pytest.raises(IndexError):
        region0.detach_block(1)


def test_op_clone():
    a = Constant.from_int_and_width(1, 32)
    b = a.clone()

    assert a is not b
    assert b.value.value.data == 1
    assert b.value.typ.width.data == 32


def test_op_clone_with_regions():
    cond = Constant.from_int_and_width(1, 1)
    a = Constant.from_int_and_width(1, 32)
    if_ = If.get(cond, [], Region.from_operation_list([a]),
                 Region.from_operation_list([a.clone()]))

    if2 = if_.clone()

    assert if2 is not if_
    assert len(if2.true_region.ops) == 1
    assert len(if2.false_region.ops) == 1
    assert if2.true_region.op is not if_.true_region.op
    assert if2.false_region.op is not if_.false_region.op


##################### Testing is_structurally_equal #####################

program_region = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 1 : !i32]
}
"""
program_region_2 = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 2 : !i32]
}
"""
program_region_2_diff_name = \
"""builtin.module() {
  %cst : !i32 = arith.constant() ["value" = 2 : !i32]
}
"""
program_region_2_diff_type = \
"""builtin.module() {
  %0 : !i64 = arith.constant() ["value" = 2 : !i64]
}
"""
program_add = \
"""builtin.module() {
%0 : !i32 = arith.constant() ["value" = 1 : !i32]
%1 : !i32 = arith.constant() ["value" = 2 : !i32]
%2 : !i32 = arith.addi(%0 : !i32, %1 : !i32)
}
"""
program_add_2 = \
"""builtin.module() {
%0 : !i32 = arith.constant() ["value" = 1 : !i32]
%1 : !i32 = arith.constant() ["value" = 2 : !i32]
%2 : !i32 = arith.addi(%1 : !i32, %0 : !i32)
}
"""
program_func = \
"""builtin.module() {
  func.func() ["sym_name" = "test", "type" = !fun<[!i32, !i32], [!i32]>, "sym_visibility" = "private"] {
  ^0(%0 : !i32, %1 : !i32):
    %2 : !i32 = arith.addi(%0 : !i32, %1 : !i32)
    func.return(%2 : !i32)
  }
}
"""
program_successors = \
"""
    func.func() ["sym_name" = "unconditional_br", "function_type" = !fun<[], []>, "sym_visibility" = "private"] {
    ^0:
        cf.br() (^1)
    ^1:
        cf.br() (^0)
    }
"""


@pytest.mark.parametrize(
    "args, expected_result",
    [([program_region, program_region], True),
     ([program_region_2, program_region_2], True),
     ([program_region_2_diff_type, program_region_2_diff_type], True),
     ([program_region_2_diff_name, program_region_2_diff_name], True),
     ([program_region, program_region_2], False),
     ([program_region_2, program_region_2_diff_type], False),
     ([program_region_2, program_region_2_diff_name], True),
     ([program_add, program_add], True),
     ([program_add_2, program_add_2], True),
     ([program_add, program_add_2], False),
     ([program_func, program_func], True),
     ([program_successors, program_successors], True),
     ([program_successors, program_func], False),
     ([program_successors, program_add], False)])
def test_is_structurally_equivalent(args: list[str], expected_result: bool):
    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_dialect(Func)
    ctx.register_dialect(Arith)
    ctx.register_dialect(Cf)

    parser = XDSLParser(ctx, args[0])
    lhs: Operation = parser.parse_op()

    parser = XDSLParser(ctx, args[1])
    rhs: Operation = parser.parse_op()

    assert lhs.is_structurally_equivalent(rhs) == expected_result


def test_is_structurally_equivalent_incompatible_ir_nodes():
    program_func = \
  """builtin.module() {
    func.func() ["sym_name" = "test", "type" = !fun<[!i32, !i32], [!i32]>, "sym_visibility" = "private"] {
    ^0(%0 : !i32, %1 : !i32):
      %2 : !i32 = arith.addi(%0 : !i32, %1 : !i32)
      %3 : !i32 = arith.constant() ["value" = 2 : !i32]
      func.return(%3 : !i32)
    ^1(%4 : !i32, %5 : !i32):
      func.return(%4 : !i32)
    }
  }
  """
    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_dialect(Func)
    ctx.register_dialect(Arith)
    ctx.register_dialect(Cf)

    parser = XDSLParser(ctx, program_func)
    program: ModuleOp = parser.parse_operation()

    assert program.is_structurally_equivalent(program.regions[0]) == False
    assert program.is_structurally_equivalent(
        program.regions[0].blocks[0]) == False
    assert program.regions[0].is_structurally_equivalent(program) == False
    assert program.regions[0].blocks[0].is_structurally_equivalent(
        program) == False
    assert program.ops[0].regions[0].blocks[0].ops[
        0].is_structurally_equivalent(
            program.ops[0].regions[0].blocks[0].ops[1]) == False
    assert program.ops[0].regions[0].blocks[0].is_structurally_equivalent(
        program.ops[0].regions[0].blocks[1]) == False



from __future__ import annotations

import re
from io import StringIO
from typing import List, Annotated

from xdsl.dialects.arith import Arith, Addi, Constant
from xdsl.dialects.builtin import Builtin, IntAttr, ModuleOp, IntegerType, UnitAttr
from xdsl.dialects.func import Func
from xdsl.ir import Attribute, MLContext, OpResult, ParametrizedAttribute, SSAValue
from xdsl.irdl import (ParameterDef, irdl_attr_definition, irdl_op_definition,
                       Operation, Operand, OptAttributeDef)
from xdsl.parser import Parser, BaseParser, Span, XDSLParser
from xdsl.printer import Printer
from xdsl.utils.diagnostic import Diagnostic


def test_simple_forgotten_op():
    """Test that the parsing of an undefined operand raises an exception."""
    ctx = MLContext()
    ctx.register_dialect(Arith)

    lit = Constant.from_int_and_width(42, 32)
    add = Addi.get(lit, lit)

    add.verify()

    expected = \
"""
%0 : !i32 = arith.addi(%<UNKNOWN> : !i32, %<UNKNOWN> : !i32)
-----------------------^^^^^^^^^^----------------------------------------------------------------
| ERROR: SSAValue is not part of the IR, are you sure all operations are added before their uses?
-------------------------------------------------------------------------------------------------
------------------------------------------^^^^^^^^^^---------------------------------------------
| ERROR: SSAValue is not part of the IR, are you sure all operations are added before their uses?
-------------------------------------------------------------------------------------------------
"""

    file = StringIO("")
    printer = Printer(stream=file)
    printer.print_op(add)

    assert file.getvalue().strip() == expected.strip()


def test_forgotten_op_non_fail():
    """Test that the parsing of an undefined operand raises an exception."""
    ctx = MLContext()
    ctx.register_dialect(Arith)

    lit = Constant.from_int_and_width(42, 32)
    add = Addi.get(lit, lit)
    add2 = Addi.get(add, add)
    mod = ModuleOp.from_region_or_ops([add, add2])
    mod.verify()

    expected = \
"""
builtin.module() {
  %0 : !i32 = arith.addi(%<UNKNOWN> : !i32, %<UNKNOWN> : !i32)
  -----------------------^^^^^^^^^^----------------------------------------------------------------
  | ERROR: SSAValue is not part of the IR, are you sure all operations are added before their uses?
  -------------------------------------------------------------------------------------------------
  ------------------------------------------^^^^^^^^^^---------------------------------------------
  | ERROR: SSAValue is not part of the IR, are you sure all operations are added before their uses?
  -------------------------------------------------------------------------------------------------
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}
"""

    file = StringIO("")
    printer = Printer(stream=file)
    printer.print_op(mod)

    assert file.getvalue().strip() == expected.strip()


@irdl_op_definition
class UnitAttrOp(Operation):
    name = "unit_attr_op"

    parallelize = OptAttributeDef(UnitAttr)


def test_unit_attr():
    """Test that a UnitAttr can be defined and printed"""

    expected = \
"""
unit_attr_op() ["parallelize"]
"""

    file = StringIO("")
    printer = Printer(stream=file)

    unit_op = UnitAttrOp.build(attributes={"parallelize": UnitAttr([])})

    printer.print_op(unit_op)
    assert file.getvalue().strip() == expected.strip()


def test_added_unit_attr():
    """Test that a UnitAttr can be added to an op, even if its not defined as a field."""

    expected = \
"""
unit_attr_op() ["parallelize", "vectorize"]
"""
    file = StringIO("")
    printer = Printer(stream=file)
    unitop = UnitAttrOp.build(attributes={
        "parallelize": UnitAttr([]),
        "vectorize": UnitAttr([])
    })

    printer.print_op(unitop)
    assert file.getvalue().strip() == expected.strip()


#  ____  _                             _   _
# |  _ \(_) __ _  __ _ _ __   ___  ___| |_(_) ___
# | | | | |/ _` |/ _` | '_ \ / _ \/ __| __| |/ __|
# | |_| | | (_| | (_| | | | | (_) \__ \ |_| | (__
# |____/|_|\__,_|\__, |_| |_|\___/|___/\__|_|\___|
#                |___/
#


def test_op_message():
    """Test that an operation message can be printed."""
    prog = \
        """builtin.module() {
    %0 : !i32 = arith.constant() ["value" = 42 : !i32]
    %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
    }"""

    expected = \
"""
builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  ^^^^^^^^^^^^^^^^^^^^^^^^^^
  | Test message
  --------------------------
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}
"""

    ctx = MLContext()
    ctx.register_dialect(Arith)
    ctx.register_dialect(Builtin)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    diagnostic = Diagnostic()
    diagnostic.add_message(module.ops[0], "Test message")
    printer = Printer(stream=file, diagnostic=diagnostic)
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_two_different_op_messages():
    """Test that an operation message can be printed."""
    prog = \
        """builtin.module() {
    %0 : !i32 = arith.constant() ["value" = 42 : !i32]
    %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
    }"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  ^^^^^^^^^^^^^^^^^^^^^^^^^^
  | Test message 1
  --------------------------
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
  ^^^^^^^^^^^^^^^^^^^^^^
  | Test message 2
  ----------------------
}"""

    ctx = MLContext()
    ctx.register_dialect(Arith)
    ctx.register_dialect(Builtin)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_module()

    file = StringIO("")
    diagnostic = Diagnostic()
    diagnostic.add_message(module.ops[0], "Test message 1")
    diagnostic.add_message(module.ops[1], "Test message 2")
    printer = Printer(stream=file, diagnostic=diagnostic)
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_two_same_op_messages():
    """Test that an operation message can be printed."""
    prog = \
        """builtin.module() {
    %0 : !i32 = arith.constant() ["value" = 42 : !i32]
    %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
    }"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  ^^^^^^^^^^^^^^^^^^^^^^^^^^
  | Test message 1
  --------------------------
  ^^^^^^^^^^^^^^^^^^^^^^^^^^
  | Test message 2
  --------------------------
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    ctx = MLContext()
    ctx.register_dialect(Arith)
    ctx.register_dialect(Builtin)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    diagnostic = Diagnostic()
    printer = Printer(stream=file, diagnostic=diagnostic)
    diagnostic.add_message(module.ops[0], "Test message 1")
    diagnostic.add_message(module.ops[0], "Test message 2")
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_op_message_with_region():
    """Test that an operation message can be printed on an operation with a region."""
    prog = \
        """builtin.module() {
    %0 : !i32 = arith.constant() ["value" = 42 : !i32]
    %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
    }"""

    expected = \
"""\
builtin.module() {
^^^^^^^^^^^^^^
| Test
--------------
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    ctx = MLContext()
    ctx.register_dialect(Arith)
    ctx.register_dialect(Builtin)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    diagnostic = Diagnostic()
    printer = Printer(stream=file, diagnostic=diagnostic)
    diagnostic.add_message(module, "Test")
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_op_message_with_region_and_overflow():
    """
    Test that an operation message can be printed on an operation with a region,
    where the message is bigger than the operation.
    """
    prog = \
        """builtin.module() {
    %0 : !i32 = arith.constant() ["value" = 42 : !i32]
    %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
    }"""

    expected = \
"""\
builtin.module() {
^^^^^^^^^^^^^^-----
| Test long message
-------------------
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    ctx = MLContext()
    ctx.register_dialect(Arith)
    ctx.register_dialect(Builtin)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    diagnostic = Diagnostic()
    printer = Printer(stream=file, diagnostic=diagnostic)
    diagnostic.add_message(module, "Test long message")
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_diagnostic():
    """
    Test that an operation message can be printed on an operation with a region,
    where the message is bigger than the operation.
    """
    prog = \
        """builtin.module() {
    %0 : !i32 = arith.constant() ["value" = 42 : !i32]
    %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
    }"""

    ctx = MLContext()
    ctx.register_dialect(Arith)
    ctx.register_dialect(Builtin)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    diag = Diagnostic()
    diag.add_message(module, "Test")
    try:
        diag.raise_exception("test message", module)
    except Exception as e:
        assert str(e)


#  ____ ____    _    _   _
# / ___/ ___|  / \  | \ | | __ _ _ __ ___   ___
# \___ \___ \ / _ \ |  \| |/ _` | '_ ` _ \ / _ \
#  ___) |__) / ___ \| |\  | (_| | | | | | |  __/
# |____/____/_/   \_\_| \_|\__,_|_| |_| |_|\___|
#


def test_print_custom_name():
    """
    Test that an SSAValue, that is a name and not a number, reserves that name
    """
    prog = \
        """builtin.module() {
    %i : !i32 = arith.constant() ["value" = 42 : !i32]
    %213 : !i32 = arith.addi(%i : !i32, %i : !i32)
    }"""

    expected = \
"""\
builtin.module() {
  %i : !i32 = arith.constant() ["value" = 42 : !i32]
  %0 : !i32 = arith.addi(%i : !i32, %i : !i32)
}"""

    ctx = MLContext()
    ctx.register_dialect(Arith)
    ctx.register_dialect(Builtin)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    printer = Printer(stream=file)
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


#   ____          _                  _____                          _
#  / ___|   _ ___| |_ ___  _ __ ___ |  ___|__  _ __ _ __ ___   __ _| |_
# | |  | | | / __| __/ _ \| '_ ` _ \| |_ / _ \| '__| '_ ` _ \ / _` | __|
# | |__| |_| \__ \ || (_) | | | | | |  _| (_) | |  | | | | | | (_| | |_
#  \____\__,_|___/\__\___/|_| |_| |_|_|  \___/|_|  |_| |_| |_|\__,_|\__|
#


@irdl_op_definition
class PlusCustomFormatOp(Operation):
    name = "test.add"
    lhs: Annotated[Operand, IntegerType]
    rhs: Annotated[Operand, IntegerType]
    res: Annotated[OpResult, IntegerType]

    @classmethod
    def parse(cls, result_types: List[Attribute],
              parser: BaseParser) -> PlusCustomFormatOp:

        lhs = parser.expect(parser.try_parse_value_id,
                            'Expected SSA Value name here!')
        parser.parse_characters("+",
                                "Malformed operation format, expected `+`!")
        rhs = parser.expect(parser.try_parse_value_id,
                            'Expected SSA Value name here!')

        return PlusCustomFormatOp.create(
            operands=[parser.get_ssa_val(lhs),
                      parser.get_ssa_val(rhs)],
            result_types=result_types)

    def print(self, printer: Printer):
        printer.print(" ", self.lhs, " + ", self.rhs)


def test_generic_format():
    """
    Test that we can use generic formats in operations.
    """
    prog = \
        """builtin.module() {
    %0 : !i32 = arith.constant() ["value" = 42 : !i32]
    %1 : !i32 = "test.add"(%0: !i32, %0: !i32)
    }"""

    expected = \
"""\
builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  %1 : !i32 = test.add %0 + %0
}"""

    ctx = MLContext()
    ctx.register_dialect(Arith)
    ctx.register_dialect(Builtin)
    ctx.register_op(PlusCustomFormatOp)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    printer = Printer(stream=file)
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_custom_format():
    """
    Test that we can use custom formats in operations.
    """
    prog = \
        """builtin.module() {
    %0 : !i32 = arith.constant() ["value" = 42 : !i32]
    %1 : !i32 = test.add %0 + %0
    }"""

    expected = \
"""\
builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  %1 : !i32 = test.add %0 + %0
}"""

    ctx = MLContext()
    ctx.register_dialect(Arith)
    ctx.register_dialect(Builtin)
    ctx.register_op(PlusCustomFormatOp)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    printer = Printer(stream=file)
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_custom_format_II():
    """
    Test that we can print using generic formats.
    """
    prog = \
        """builtin.module() {
    %0 : !i32 = arith.constant() ["value" = 42 : !i32]
    %1 : !i32 = test.add %0 + %0
    }"""

    expected = \
"""\
"builtin.module"() {
  %0 : !i32 = "arith.constant"() ["value" = 42 : !i32]
  %1 : !i32 = "test.add"(%0 : !i32, %0 : !i32)
}"""

    ctx = MLContext()
    ctx.register_dialect(Arith)
    ctx.register_dialect(Builtin)
    ctx.register_op(PlusCustomFormatOp)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    printer = Printer(stream=file, print_generic_format=True)
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


@irdl_attr_definition
class CustomFormatAttr(ParametrizedAttribute):
    name = "custom"

    attr: ParameterDef[IntAttr]

    @staticmethod
    def parse_parameters(parser: BaseParser) -> list[Attribute]:
        parser.parse_char("<")
        value = parser.tokenizer.next_token_of_pattern(
            re.compile('(zero|one)'))
        if value and value.text == "zero":
            parser.parse_char(">")
            return [IntAttr.from_int(0)]
        if value and value.text == "one":
            parser.parse_char(">")
            return [IntAttr.from_int(1)]
        assert False

    def print_parameters(self, printer: Printer) -> None:
        assert 0 <= self.attr.data <= 1
        printer.print("<", "zero" if self.attr.data == 0 else "one", ">")


@irdl_op_definition
class AnyOp(Operation):
    name = "any"


def test_custom_format_attr():
    """
    Test that we can parse and print attributes using custom formats.
    """
    prog = \
        """builtin.module() {
      any() ["attr" = !custom<zero>]
    }"""

    expected = \
"""\
builtin.module() {
  any() ["attr" = !custom<zero>]
}"""

    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_op(AnyOp)
    ctx.register_attr(CustomFormatAttr)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    printer = Printer(stream=file)
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_parse_generic_format_attr():
    """
    Test that we can parse attributes using generic formats.
    """
    prog = \
        """builtin.module() {
      any() ["attr" = #"custom"<#int<0>>]
    }"""

    expected = \
"""\
builtin.module() {
  any() ["attr" = !custom<zero>]
}"""

    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_op(AnyOp)
    ctx.register_attr(CustomFormatAttr)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    printer = Printer(stream=file)
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_parse_generic_format_attr_II():
    """
    Test that we can parse attributes using generic formats.
    """
    prog = \
        """builtin.module() {
      any() ["attr" = !custom<zero>]
    }"""

    expected = \
"""\
"builtin.module"() {
  "any"() ["attr" = !"custom"<!int<0>>]
}"""

    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_op(AnyOp)
    ctx.register_attr(CustomFormatAttr)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    printer = Printer(stream=file, print_generic_format=True)
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_parse_generic_format_attr_III():
    """
    Test that we can parse attributes using generic formats.
    """
    prog = \
        """builtin.module() {
      any() ["attr" = !custom<one>]
    }"""

    expected = \
"""\
"builtin.module"() {
  "any"() ["attr" = !"custom"<!int<1>>]
}"""

    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_op(AnyOp)
    ctx.register_attr(CustomFormatAttr)

    parser = Parser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    printer = Printer(stream=file, print_generic_format=True)
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_parse_dense_xdsl():
    """
    Test that parsing of shaped dense tensors works.
    """
    # TODO: handle nested array syntax
    prog = """
    %0 : !tensor<[2 : !index, 3 : !index], !f64> = arith.constant() ["value" = !dense<!tensor<[2 : !index, 3 : !index], !f64>, [1.0 : !f64, 2.0 : !f64, 3.0 : !f64, 4.0 : !f64, 5.0 : !f64, 6.0 : !f64]>]
    """

    expected = """
    %0 : !tensor<[2 : !index, 3 : !index], !f64> = arith.constant() ["value" = !dense<!tensor<[2 : !index, 3 : !index], !f64>, [1.0 : !f64, 2.0 : !f64, 3.0 : !f64, 4.0 : !f64, 5.0 : !f64, 6.0 : !f64]>]
    """

    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_dialect(Arith)

    parser = XDSLParser(ctx, prog)
    module = parser.parse_op()

    file = StringIO("")
    printer = Printer(stream=file)
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_parse_dense_mlir():
    """
    Test that we can parse attributes using generic formats.
    """
    prog = """
    %0 = "arith.constant"() {"value" = dense<[[0.0, 1.0, 2.0], [3.0, 4.0, 5.0]]> : tensor<2x3xf64>} : () -> tensor<2x3xf64>
    """

    expected = """
    %0 = "arith.constant"() {"value" = dense<[[0.0, 1.0, 2.0], [3.0, 4.0, 5.0]]> : tensor<2x3xf64>} : () -> tensor<2x3xf64>
    """

    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_dialect(Arith)

    parser = Parser(ctx, prog, source=Parser.Source.MLIR)
    module = parser.parse_op()

    file = StringIO("")
    printer = Printer(stream=file, target=Printer.Target.MLIR)
    printer.print_op(module)
    assert file.getvalue().strip() == expected.strip()


def test_foo_string():
    """
    Fail attribute in purpose.
    """
    prog = \
        """builtin.module() {
      any() ["attr" = !"string"<"foo">]
    }"""

    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_op(AnyOp)
    ctx.register_attr(CustomFormatAttr)

    parser = XDSLParser(ctx, prog)
    try:
        parser.parse_op()
        assert False
    except:
        pass


def test_dictionary_attr():
    """Test that a DictionaryAttr can be parsed and then printed."""

    prog = """
    func.func() ["sym_name" = "test", "function_type" = !i64, "sym_visibility" = "private", "arg_attrs" = {"key_one"="value_one", "key_two"="value_two", "key_three"=72 : !i64}]
    """

    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_dialect(Func)

    parser = XDSLParser(ctx, prog)
    parsed = parser.parse_op()

    file = StringIO("")
    printer = Printer(stream=file)
    printer.print_op(parsed)
    assert file.getvalue().strip() == prog.strip()



from xdsl.ir import MLContext, Use
from xdsl.dialects.builtin import ModuleOp, Builtin
from xdsl.parser import Parser
from xdsl.dialects.arith import Arith

test_prog = """
builtin.module() {
  %0 : !i1 = arith.constant() ["value" = 0 : !i1]
  %1 : !i1 = arith.andi(%0 : !i1, %0 : !i1)
}
"""


def test_main():
    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_dialect(Arith)

    parser = Parser(ctx, test_prog)
    module = parser.parse_op()

    module.verify()
    assert isinstance(module, ModuleOp)
    constant_op = module.ops[0]
    andi_op = module.ops[1]
    assert constant_op.results[0].uses == {Use(andi_op, 0), Use(andi_op, 1)}
    assert andi_op.results[0].uses == set()

    print("Done")



from io import StringIO
from typing import Callable

from xdsl.dialects.arith import Arith, Constant, Addi
from xdsl.dialects.builtin import ModuleOp, Builtin, i32
from xdsl.dialects.scf import Scf, Yield
from xdsl.dialects.func import Func
from xdsl.ir import MLContext, Block
from xdsl.parser import Parser
from xdsl.printer import Printer
from xdsl.rewriter import Rewriter


def rewrite_and_compare(prog: str, expected_prog: str,
                        transformation: Callable[[ModuleOp, Rewriter], None]):
    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_dialect(Arith)
    ctx.register_dialect(Scf)
    ctx.register_dialect(Func)

    parser = Parser(ctx, prog)
    module = parser.parse_op()

    rewriter = Rewriter()
    transformation(module, rewriter)
    file = StringIO("")
    printer = Printer(stream=file)
    printer.print_op(module)
    assert file.getvalue().strip() == expected_prog.strip()


def test_operation_deletion():
    """Test rewrites where SSA values are deleted."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    expected = \
"""builtin.module() {}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        constant_op = module.ops[0]
        rewriter.erase_op(constant_op)

    rewrite_and_compare(prog, expected, transformation)


# Test an operation replacement
def test_replace_op_one_op():
    prog = \
    """builtin.module() {
%0 : !i32 = arith.constant() ["value" = 42 : !i32]
%1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 43 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        constant_op = module.ops[0]
        new_constant_op = Constant.from_int_and_width(43, i32)
        rewriter.replace_op(constant_op, new_constant_op)

    rewrite_and_compare(prog, expected, transformation)


# Test an operation replacement with multiple ops
def test_replace_op_multiple_op():
    prog = \
    """builtin.module() {
%0 : !i32 = arith.constant() ["value" = 2 : !i32]
%1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 1 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
  %2 : !i32 = arith.addi(%1 : !i32, %1 : !i32)
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        constant_op = module.ops[0]
        new_constant = Constant.from_int_and_width(1, i32)
        new_add = Addi.get(new_constant, new_constant)

        rewriter.replace_op(constant_op, [new_constant, new_add])

    rewrite_and_compare(prog, expected, transformation)


# Test an operation replacement with manually specified results
def test_replace_op_new_results():
    prog = \
    """builtin.module() {
%0 : !i32 = arith.constant() ["value" = 2 : !i32]
%1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
%2 : !i32 = arith.muli(%1 : !i32, %1 : !i32)
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 2 : !i32]
  %1 : !i32 = arith.muli(%0 : !i32, %0 : !i32)
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        add_op = module.ops[1]

        rewriter.replace_op(add_op, [], [add_op.lhs])

    rewrite_and_compare(prog, expected, transformation)


def test_inline_block_at_pos():
    """Test the inlining of a block at a certain position."""
    prog = \
    """builtin.module() {
%0 : !i1 = arith.constant() ["value" = true]
scf.if(%0 : !i1) {
  %1 : !i32 = arith.constant() ["value" = 2 : !i32]
}
}"""

    expected = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  %1 : !i32 = arith.constant() ["value" = 2 : !i32]
  scf.if(%0 : !i1) {}
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        if_op = module.ops[1]
        module_block = module.regions[0].blocks[0]
        if_block = if_op.regions[0].blocks[0]

        rewriter.inline_block_at_pos(if_block, module_block, 1)

    rewrite_and_compare(prog, expected, transformation)


def test_inline_block_before():
    """Test the inlining of a block before an operation."""
    prog = \
    """builtin.module() {
%0 : !i1 = arith.constant() ["value" = true]
scf.if(%0 : !i1) {
  %1 : !i32 = arith.constant() ["value" = 2 : !i32]
}
}"""

    expected = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  %1 : !i32 = arith.constant() ["value" = 2 : !i32]
  scf.if(%0 : !i1) {}
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        if_op = module.ops[1]
        if_block = if_op.regions[0].blocks[0]

        rewriter.inline_block_before(if_block, if_op)

    rewrite_and_compare(prog, expected, transformation)


def test_inline_block_after():
    """Test the inlining of a block after an operation."""
    prog = \
    """builtin.module() {
%0 : !i1 = arith.constant() ["value" = true]
scf.if(%0 : !i1) {
  %1 : !i32 = arith.constant() ["value" = 2 : !i32]
}
}"""

    expected = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  %1 : !i32 = arith.constant() ["value" = 2 : !i32]
  scf.if(%0 : !i1) {}
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        if_op = module.ops[1]
        constant_op = module.ops[0]
        if_block = if_op.regions[0].blocks[0]

        rewriter.inline_block_after(if_block, constant_op)

    rewrite_and_compare(prog, expected, transformation)


def test_insert_block():
    """Test the insertion of a block in a region."""
    prog = \
    """builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
}"""

    expected = \
"""builtin.module() {
^0:
^1:
  %0 : !i1 = arith.constant() ["value" = true]
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        module.regions[0].insert_block(Block(), 0)

    rewrite_and_compare(prog, expected, transformation)


def test_insert_block2():
    """Test the insertion of a block in a region."""
    prog = \
    """builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
}"""

    expected = \
"""builtin.module() {
^0:
  %0 : !i1 = arith.constant() ["value" = true]
^1:
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        module.regions[0].insert_block(Block(), 1)

    rewrite_and_compare(prog, expected, transformation)


def test_insert_block_before():
    """Test the insertion of a block before another block."""
    prog = \
    """builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
}"""

    expected = \
"""builtin.module() {
^0:
^1:
  %0 : !i1 = arith.constant() ["value" = true]
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        rewriter.insert_block_before(Block(), module.regions[0].blocks[0])

    rewrite_and_compare(prog, expected, transformation)


def test_insert_block_after():
    """Test the insertion of a block after another block."""
    prog = \
    """builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
}"""

    expected = \
"""builtin.module() {
^0:
  %0 : !i1 = arith.constant() ["value" = true]
^1:
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        rewriter.insert_block_after(Block(), module.regions[0].blocks[0])

    rewrite_and_compare(prog, expected, transformation)


def test_preserve_naming_single_op():
    """Test the preservation of names of SSAValues"""
    prog = \
    """builtin.module() {
   %i : !i32 = arith.constant() ["value" = 42 : !i32]
    %1 : !i32 = arith.addi(%i : !i32, %i : !i32)
}"""

    expected = \
"""builtin.module() {
  %i : !i32 = arith.constant() ["value" = 1 : !i32]
  %0 : !i32 = arith.addi(%i : !i32, %i : !i32)
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        constant_op = module.ops[0]
        new_constant = Constant.from_int_and_width(1, i32)

        rewriter.replace_op(constant_op, [new_constant])

    rewrite_and_compare(prog, expected, transformation)


def test_preserve_naming_multiple_ops():
    """Test the preservation of names of SSAValues for transformations to multiple ops"""
    prog = \
    """builtin.module() {
   %i : !i32 = arith.constant() ["value" = 42 : !i32]
    %1 : !i32 = arith.addi(%i : !i32, %i : !i32)
}"""

    expected = \
"""builtin.module() {
  %i : !i32 = arith.constant() ["value" = 1 : !i32]
  %i1 : !i32 = arith.addi(%i : !i32, %i : !i32)
  %0 : !i32 = arith.addi(%i1 : !i32, %i1 : !i32)
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        constant_op = module.ops[0]
        new_constant = Constant.from_int_and_width(1, i32)
        new_add = Addi.get(new_constant, new_constant)

        rewriter.replace_op(constant_op, [new_constant, new_add])

    rewrite_and_compare(prog, expected, transformation)


def test_no_result_rewriter():
    """Test rewriter on ops without results"""
    prog = \
    """builtin.module() {
   func.return()
}"""

    expected = \
"""builtin.module() {
  scf.yield()
}"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        return_op = module.ops[0]
        new_op = Yield.get()

        rewriter.replace_op(return_op, [new_op])

    rewrite_and_compare(prog, expected, transformation)



from __future__ import annotations
from typing import Annotated
import pytest

from xdsl.dialects.builtin import (DenseIntOrFPElementsAttr, VectorType,
                                   IntegerType, Operation, StringAttr, i32)
from xdsl.dialects.arith import Constant

from xdsl.ir import Block, OpResult, Region
from xdsl.irdl import (OptOpResult, OptOperand, OptRegion,
                       OptSingleBlockRegion, Operand, SingleBlockRegion,
                       VarOpResult, VarRegion, VarSingleBlockRegion,
                       irdl_op_definition, AttrSizedResultSegments, VarOperand,
                       AttrSizedOperandSegments, OpAttr, Region, OptOpAttr)

#  ____                 _ _
# |  _ \ ___  ___ _   _| | |_
# | |_) / _ \/ __| | | | | __|
# |  _ <  __/\__ \ |_| | | |_
# |_| \_\___||___/\__,_|_|\__|
#


@irdl_op_definition
class ResultOp(Operation):
    name: str = "test.result_op"

    res: Annotated[OpResult, StringAttr]


def test_result_builder():
    op = ResultOp.build(result_types=[0])
    op.verify()
    assert [res.typ for res in op.results] == [StringAttr.from_int(0)]


def test_result_builder_exception():
    with pytest.raises(ValueError):
        ResultOp.build()


@irdl_op_definition
class OptResultOp(Operation):
    name: str = "test.opt_result_op"

    res: Annotated[OptOpResult, StringAttr]


def test_opt_result_builder():
    op1 = OptResultOp.build(result_types=[[0]])
    op2 = OptResultOp.build(result_types=[[]])
    op1.verify()
    op2.verify()
    assert [res.typ for res in op1.results] == [StringAttr.from_int(0)]
    assert len(op2.results) == 0


def test_opt_result_builder_two_args():
    with pytest.raises(ValueError) as _:
        OptResultOp.build(result_types=[[0, 0]])


@irdl_op_definition
class VarResultOp(Operation):
    name: str = "test.var_result_op"

    res: Annotated[VarOpResult, StringAttr]


def test_var_result_builder():
    op = VarResultOp.build(result_types=[[0, 1]])
    op.verify()
    assert [res.typ for res in op.results
            ] == [StringAttr.from_int(0),
                  StringAttr.from_int(1)]


@irdl_op_definition
class TwoVarResultOp(Operation):
    name: str = "test.two_var_result_op"

    res1: Annotated[VarOpResult, StringAttr]
    res2: Annotated[VarOpResult, StringAttr]
    irdl_options = [AttrSizedResultSegments()]


def test_two_var_result_builder():
    op = TwoVarResultOp.build(result_types=[[0, 1], [2, 3]])
    op.verify()
    assert [res.typ for res in op.results] == [
        StringAttr.from_int(0),
        StringAttr.from_int(1),
        StringAttr.from_int(2),
        StringAttr.from_int(3)
    ]

    dense_type = VectorType.from_element_type_and_shape(
        IntegerType.from_width(32), [2])

    assert op.attributes[AttrSizedResultSegments.
                         attribute_name] == DenseIntOrFPElementsAttr.from_list(
                             dense_type, [2, 2])


def test_two_var_result_builder2():
    op = TwoVarResultOp.build(result_types=[[0], [1, 2, 3]])
    op.verify()
    assert [res.typ for res in op.results] == [
        StringAttr.from_int(0),
        StringAttr.from_int(1),
        StringAttr.from_int(2),
        StringAttr.from_int(3)
    ]
    dense_type = VectorType.from_element_type_and_shape(
        IntegerType.from_width(32), [2])
    assert op.attributes[AttrSizedResultSegments.
                         attribute_name] == DenseIntOrFPElementsAttr.from_list(
                             dense_type, [1, 3])


@irdl_op_definition
class MixedResultOp(Operation):
    name: str = "test.mixed"

    res1: Annotated[VarOpResult, StringAttr]
    res2: Annotated[OpResult, StringAttr]
    res3: Annotated[VarOpResult, StringAttr]
    irdl_options = [AttrSizedResultSegments()]


def test_var_mixed_builder():
    op = MixedResultOp.build(result_types=[[0, 1], 2, [3, 4]])
    op.verify()
    assert [res.typ for res in op.results] == [
        StringAttr.from_int(0),
        StringAttr.from_int(1),
        StringAttr.from_int(2),
        StringAttr.from_int(3),
        StringAttr.from_int(4)
    ]

    dense_type = VectorType.from_element_type_and_shape(
        IntegerType.from_width(32), [3])
    assert op.attributes[AttrSizedResultSegments.
                         attribute_name] == DenseIntOrFPElementsAttr.from_list(
                             dense_type, [2, 1, 2])


#   ___                                 _
#  / _ \ _ __   ___ _ __ __ _ _ __   __| |
# | | | | '_ \ / _ \ '__/ _` | '_ \ / _` |
# | |_| | |_) |  __/ | | (_| | | | | (_| |
#  \___/| .__/ \___|_|  \__,_|_| |_|\__,_|
#       |_|
#


@irdl_op_definition
class OperandOp(Operation):
    name: str = "test.operand_op"

    res: Annotated[Operand, StringAttr]


def test_operand_builder_operation():
    op1 = ResultOp.build(result_types=[0])
    op2 = OperandOp.build(operands=[op1])
    op2.verify()
    assert op2.operands == [op1.res]


def test_operand_builder_value():
    op1 = ResultOp.build(result_types=[0])
    op2 = OperandOp.build(operands=[op1.res])
    op2.verify()
    assert op2.operands == [op1.res]


def test_operand_builder_exception():
    with pytest.raises(ValueError):
        OperandOp.build()


@irdl_op_definition
class OptOperandOp(Operation):
    name: str = "test.opt_operand_op"

    res: Annotated[OptOperand, StringAttr]


def test_opt_operand_builder():
    op = ResultOp.build(result_types=[0])
    op1 = OptOperandOp.build(operands=[[op]])
    op2 = OptOperandOp.build(operands=[[]])
    op1.verify()
    op2.verify()
    assert [operand for operand in op1.operands] == [op.res]
    assert len(op2.operands) == 0


def test_opt_operand_builder_two_args():
    op = ResultOp.build(result_types=[0])
    with pytest.raises(ValueError) as _:
        OptOperandOp.build(operands=[[op, op]])


@irdl_op_definition
class VarOperandOp(Operation):
    name: str = "test.var_operand_op"

    res: Annotated[VarOperand, StringAttr]


def test_var_operand_builder():
    op1 = ResultOp.build(result_types=[0])
    op2 = VarOperandOp.build(operands=[[op1, op1]])
    op2.verify()
    assert op2.operands == [op1.res, op1.res]


@irdl_op_definition
class TwoVarOperandOp(Operation):
    name: str = "test.two_var_operand_op"

    res1: Annotated[VarOperand, StringAttr]
    res2: Annotated[VarOperand, StringAttr]
    irdl_options = [AttrSizedOperandSegments()]


def test_two_var_operand_builder():
    op1 = ResultOp.build(result_types=[0])
    op2 = TwoVarOperandOp.build(operands=[[op1, op1], [op1, op1]])
    op2.verify()
    assert op2.operands == [op1.res] * 4
    dense_type = VectorType.from_element_type_and_shape(
        IntegerType.from_width(32), [2])
    assert op2.attributes[
        AttrSizedOperandSegments.
        attribute_name] == DenseIntOrFPElementsAttr.from_list(
            dense_type, [2, 2])


def test_two_var_operand_builder2():
    op1 = ResultOp.build(result_types=[0])
    op2 = TwoVarOperandOp.build(operands=[[op1], [op1, op1, op1]])
    op2.verify()
    assert op2.operands == [op1.res] * 4
    dense_type = VectorType.from_element_type_and_shape(
        IntegerType.from_width(32), [2])
    assert op2.attributes[
        AttrSizedOperandSegments.
        attribute_name] == DenseIntOrFPElementsAttr.from_list(
            dense_type, [1, 3])


#      _   _   _        _ _           _
#     / \ | |_| |_ _ __(_) |__  _   _| |_ ___
#    / _ \| __| __| '__| | '_ \| | | | __/ _ \
#   / ___ \ |_| |_| |  | | |_) | |_| | ||  __/
#  /_/   \_\__|\__|_|  |_|_.__/ \__,_|\__\___|
#


@irdl_op_definition
class AttrOp(Operation):
    name: str = "test.two_var_result_op"
    attr: OpAttr[StringAttr]


def test_attr_op():
    op = AttrOp.build(attributes={"attr": 0})
    op.verify()
    assert op.attr == StringAttr.from_int(0)


def test_attr_new_attr_op():
    op = AttrOp.build(attributes={
        "attr": 0,
        "new_attr": StringAttr.from_int(1)
    })
    op.verify()
    assert op.attr == StringAttr.from_int(0)
    assert op.attributes["new_attr"] == StringAttr.from_int(1)


@irdl_op_definition
class OptionalAttrOp(Operation):
    name: str = "test.opt_attr_op"

    opt_attr: OptOpAttr[StringAttr]


def test_optional_attr_op_empty():
    op = OptionalAttrOp.build()
    op.verify()
    assert op.opt_attr is None


#  ____            _
# |  _ \ ___  __ _(_) ___  _ __
# | |_) / _ \/ _` | |/ _ \| '_ \
# |  _ <  __/ (_| | | (_) | | | |
# |_| \_\___|\__, |_|\___/|_| |_|
#            |___/
#


@irdl_op_definition
class RegionOp(Operation):
    name: str = "test.region_op"

    region: Region


def test_region_op_region():
    op = RegionOp.build(regions=[Region()])
    op.verify()

    assert op.region.blocks == []


def test_region_op_blocks():
    op = RegionOp.build(regions=[[Block(), Block()]])
    op.verify()
    assert len(op.region.blocks) == 2


def test_region_op_ops():
    op1 = RegionOp.build(regions=[[Block()]])
    op2 = RegionOp.build(regions=[[Block()]])
    op = RegionOp.build(regions=[[op1, op2]])
    op.verify()
    assert len(op.region.blocks) == 1
    assert len(op.region.blocks[0].ops) == 2


def test_noop_region():
    region0 = Region.get([])
    assert len(region0.ops) == 0


def test_singleop_region():
    a = Constant.from_int_and_width(1, i32)
    region0 = Region.get([a])
    assert type(region0.op) is Constant


@irdl_op_definition
class SBRegionOp(Operation):
    name: str = "test.sbregion_op"

    region: SingleBlockRegion


def test_sbregion_one_block():
    op = SBRegionOp.build(regions=[[Block()]])
    op.verify()
    assert len(op.region.blocks) == 1


@irdl_op_definition
class OptRegionOp(Operation):
    name: str = "test.opt_region_op"

    reg: OptRegion


def test_opt_region_builder():
    op1 = OptRegionOp.build(regions=[[[Block(), Block()]]])
    op2 = OptRegionOp.build(regions=[[Region()]])
    op1.verify()
    op2.verify()


def test_opt_region_builder_two_args():
    with pytest.raises(ValueError) as _:
        OptRegionOp.build(regions=[[Region(), Region()]])


@irdl_op_definition
class OptSBRegionOp(Operation):
    name: str = "test.sbregion_op"

    region: OptSingleBlockRegion


def test_opt_sbregion_one_block():
    op1 = OptSBRegionOp.build(regions=[[[Block()]]])
    op2 = OptSBRegionOp.build(regions=[[]])
    op1.verify()
    op2.verify()
    assert len(op1.region.blocks) == 1
    assert op2.region is None


@irdl_op_definition
class VarRegionOp(Operation):
    name: str = "test.var_operand_op"

    regs: VarRegion


def test_var_region_builder():
    op = VarRegionOp.build(regions=[[Region(), [Block(), Block()]]])
    op.verify()
    assert len(op.regs[0].blocks) == 0  # type: ignore
    assert len(op.regs[1].blocks) == 2  # type: ignore


@irdl_op_definition
class VarSBRegionOp(Operation):
    name: str = "test.sbregion_op"

    regs: VarSingleBlockRegion


def test_var_sbregion_one_block():
    op1 = VarSBRegionOp.build(regions=[[[Block()]]])
    op2 = VarSBRegionOp.build(regions=[[Region(), [Block(), Block()]]])
    op1.verify()
    op2.verify()
    assert len(op1.regs) == 1  # type: ignore
    assert len(op2.regs) == 2  # type: ignore
    assert len(op2.regs[0].blocks) == 0  # type: ignore
    assert len(op2.regs[1].blocks) == 2  # type: ignore


#  __  __ _
# |  \/  (_)___  ___
# | |\/| | / __|/ __|
# | |  | | \__ \ (__
# |_|  |_|_|___/\___|
#


def test_parent_pointers():
    op = ResultOp.build(result_types=[0])
    block = Block.from_ops([op])
    reg = Region.from_block_list([block])
    reg_op = RegionOp.build(regions=[reg])

    block_2 = Block.from_ops([reg_op])
    reg_2 = Region.from_block_list([block_2])
    reg_op_2 = RegionOp.build(regions=[reg_2])

    assert op.parent_block() is block
    assert op.parent_region() is reg
    assert op.parent_op() is reg_op

    assert reg_op_2.parent_block() is None
    assert reg_op_2.parent_region() is None
    assert reg_op_2.parent_op() is None

    assert reg.parent_op() is reg_op
    assert reg.parent_block() is block_2
    assert reg.parent_region() is reg_2

    assert reg_2.parent_block() is None
    assert reg_2.parent_region() is None

    assert block.parent_region() is reg
    assert block.parent_op() is reg_op
    assert block.parent_block() is block_2

    assert block_2.parent_block() is None



from __future__ import annotations
from typing import cast

from xdsl.dialects.builtin import AnyArrayAttr, ArrayAttr, StringAttr
from xdsl.ir import (ParametrizedAttribute, Operation, Attribute, Dialect)
from xdsl.irdl import (ParameterDef, irdl_op_definition, irdl_attr_definition,
                       SingleBlockRegion, OpAttr)
from xdsl.parser import BaseParser
from xdsl.printer import Printer


@irdl_attr_definition
class EqTypeConstraintAttr(ParametrizedAttribute):

    name = "irdl.equality_type_constraint"
    type: ParameterDef[Attribute]


@irdl_attr_definition
class AnyTypeConstraintAttr(ParametrizedAttribute):
    name = "irdl.any_type_constraint"


@irdl_attr_definition
class AnyOfTypeConstraintAttr(ParametrizedAttribute):
    name = "irdl.any_of_type"
    params: ParameterDef[ArrayAttr[Attribute]]


@irdl_attr_definition
class VarTypeConstraintAttr(ParametrizedAttribute):
    name = "irdl.var_type_constraint"
    var_name: ParameterDef[StringAttr]


@irdl_attr_definition
class DynTypeBaseConstraintAttr(ParametrizedAttribute):
    name = "irdl.dyn_type_constraint"
    type_name: ParameterDef[StringAttr]


@irdl_attr_definition
class DynTypeParamsConstraintAttr(ParametrizedAttribute):
    name = "irdl.dyn_type_params_constraint"
    type_name: ParameterDef[StringAttr]
    params: ParameterDef[AnyArrayAttr]


@irdl_attr_definition
class TypeParamsConstraintAttr(ParametrizedAttribute):
    name = "irdl.type_params_constraint"
    type_name: ParameterDef[StringAttr]
    params: ParameterDef[AnyArrayAttr]


@irdl_attr_definition
class NamedTypeConstraintAttr(ParametrizedAttribute):
    name = "irdl.named_type_constraint"
    type_name: ParameterDef[StringAttr]
    params_constraints: ParameterDef[Attribute]

    @staticmethod
    def parse_parameters(parser: BaseParser) -> list[Attribute]:
        parser.parse_char("<")
        type_name = parser.parse_str_literal()
        parser.parse_char(":")
        params_constraints = parser.parse_attribute()
        parser.parse_char(">")
        return [StringAttr.from_str(type_name), params_constraints]

    def print_parameters(self, printer: Printer) -> None:
        printer.print("<\"", self.type_name.data, "\" : ",
                      self.params_constraints, ">")


@irdl_op_definition
class DialectOp(Operation):
    """
    Define a new dialect
    """
    name = "irdl.dialect"
    body: SingleBlockRegion

    @property
    def dialect_name(self) -> StringAttr:
        return cast(StringAttr, self.attributes["name"])

    def verify_(self) -> None:
        if "name" not in self.attributes.keys():
            raise ValueError("name attribute is required")
        if not isinstance(self.attributes["name"], StringAttr):
            raise ValueError("name attribute must be a string attribute")

    def get_op_defs(self) -> list[OperationOp]:
        """Get the operations defined by the dialect"""
        return [op for op in self.body.ops if isinstance(op, OperationOp)]

    def get_type_defs(self) -> list[TypeOp]:
        """Get the types defined by the dialect"""
        return [op for op in self.body.ops if isinstance(op, TypeOp)]


@irdl_op_definition
class ParametersOp(Operation):
    """
    Define the parameters of a type/attribute definition
    """
    name = "irdl.parameters"
    params: OpAttr[AnyArrayAttr]


@irdl_op_definition
class TypeOp(Operation):
    """
    Defines new types belonging to previously defined dialect
    """
    name = "irdl.type"
    body: SingleBlockRegion

    @property
    def type_name(self) -> StringAttr:
        return cast(StringAttr, self.attributes["name"])

    def verify_(self) -> None:
        if "name" not in self.attributes.keys():
            raise ValueError("name attribute is required")
        if not isinstance(self.attributes["name"], StringAttr):
            raise ValueError("name attribute must be a string attribute")


@irdl_op_definition
class ConstraintVarsOp(Operation):
    """
    Define constraint variables that can be used in the
    current region
    """
    name = "irdl.constraint_vars"
    constraints: OpAttr[Attribute]


@irdl_op_definition
class OperandsOp(Operation):
    """
    Define the operands of a parent operation
    """
    name = "irdl.operands"
    params: OpAttr[Attribute]


@irdl_op_definition
class ResultsOp(Operation):
    """
    Define results of parent operation
    """
    name = "irdl.results"
    params: OpAttr[Attribute]


@irdl_op_definition
class OperationOp(Operation):
    """
    Define a new operation belonging to previously defined dialect
    """
    name = "irdl.operation"
    body: SingleBlockRegion

    @property
    def op_name(self) -> StringAttr:
        return cast(StringAttr, self.attributes["name"])

    def verify_(self) -> None:
        if "name" not in self.attributes.keys():
            raise ValueError("name attribute is required")
        if not isinstance(self.attributes["name"], StringAttr):
            raise ValueError("name attribute must be a string attribute")

    def get_operands(self) -> OperandsOp | None:
        """Get the operation operands definition"""
        for op in self.body.ops:
            if isinstance(op, OperandsOp):
                return op
        return None

    def get_results(self) -> ResultsOp | None:
        """Get the operation results definition"""
        for op in self.body.ops:
            if isinstance(op, ResultsOp):
                return op
        return None


IRDL = Dialect(
    [
        DialectOp,
        ParametersOp,
        TypeOp,
        ConstraintVarsOp,
        OperandsOp,
        ResultsOp,
        OperationOp,
    ],
    [
        AnyTypeConstraintAttr,
        AnyOfTypeConstraintAttr,  #
        EqTypeConstraintAttr,
        VarTypeConstraintAttr,
        TypeParamsConstraintAttr,
        NamedTypeConstraintAttr,
        DynTypeBaseConstraintAttr,
        DynTypeParamsConstraintAttr
    ])



_mlir_module = None
_REQUIRED_MLIR_MODULES = ['ir']

import types as _xdsl_init_types

# pyright: reportMissingImports=false, reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false


def load_mlir_module(mlir_module: _xdsl_init_types.ModuleType) -> None:
    """
    Configures xDSL to use the given MLIR module. This allows to use xDSL with
    out-of-tree dialects, which ship with their own copy of the MLIR core
    modules (see https://github.com/xdslproject/xdsl/issues/111). If so desired,
    the user needs to call this function prior to importing `mlir_converter`.
    Otherwise, `mlir` is used. Subsequent calls to this function are only
    allowed if they are made with the same argument. Furthermore, the provided
    module needs to be imported including all submodules specified in
    _REQUIRED_MLIR_MODULES by the caller prior to calling this function. If we
    load_mlir_module is called the first time (i.e. when _mlir_module is still
    None), the given argument is written into _mlir_module (given that the
    required submodules are loaded). If the function is called again, the given
    mlir_module argument needs to be the same each time.
    """
    global _mlir_module
    if _mlir_module and _mlir_module != mlir_module:
        raise RuntimeError('Different modules already loaded previously.')
    for submodule in _REQUIRED_MLIR_MODULES:
        if not hasattr(mlir_module, submodule):
            raise RuntimeError(
                'Provided module "{}" does not have the submodule "{}" loaded.'
                .format(mlir_module.__name__, submodule))
    _mlir_module = mlir_module


def ensure_mlir_module_loaded() -> None:
    """
    Loads the default `mlir` module if no module has been loaded previously by
    `load_mlir_module`.
    """
    if _mlir_module:
        return
    import mlir
    import mlir.ir
    load_mlir_module(mlir)



from __future__ import annotations
from typing import Annotated

from xdsl.ir import Attribute, OpResult, Operation, Region
from xdsl.irdl import (Operand, irdl_op_definition, OperandDef, ResultDef,
                       AttributeDef, AnyAttr, OpDef, RegionDef, OpAttr)

#  ___ ____  ____  _     ____        __
# |_ _|  _ \|  _ \| |   |  _ \  ___ / _|
#  | || |_) | | | | |   | | | |/ _ \ |_
#  | ||  _ <| |_| | |___| |_| |  __/  _|
# |___|_| \_\____/|_____|____/ \___|_|
#


@irdl_op_definition
class OpDefTestOp(Operation):
    name = "test.op_def_test"

    operand: Operand
    result: OpResult
    attr: OpAttr[Attribute]
    region: Region


def test_get_definition():
    """Test retrieval of an IRDL definition from an operation"""
    assert OpDefTestOp.irdl_definition == OpDef(
        "test.op_def_test",
        operands=[("operand", OperandDef(AnyAttr()))],
        results=[("result", ResultDef(AnyAttr()))],
        attributes={"attr": AttributeDef(AnyAttr())},
        regions=[("region", RegionDef())])



from io import StringIO

from xdsl.dialects.arith import Arith, Constant, Addi, Muli
from xdsl.dialects.builtin import i32, i64, Builtin, IntegerAttr, ModuleOp
from xdsl.dialects.scf import If, Scf
from xdsl.ir import MLContext, Region, Operation
from xdsl.pattern_rewriter import (PatternRewriteWalker,
                                   op_type_rewrite_pattern, RewritePattern,
                                   PatternRewriter, AnonymousRewritePattern,
                                   GreedyRewritePatternApplier)
from xdsl.parser import Parser
from xdsl.printer import Printer


def rewrite_and_compare(prog: str, expected_prog: str,
                        walker: PatternRewriteWalker):
    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_dialect(Arith)
    ctx.register_dialect(Scf)

    parser = Parser(ctx, prog)
    module = parser.parse_op()

    assert isinstance(module, ModuleOp)

    walker.rewrite_module(module)
    file = StringIO("")
    printer = Printer(stream=file)
    printer.print_op(module)
    assert file.getvalue().strip() == expected_prog.strip()


def test_non_recursive_rewrite():
    """Test a simple non-recursive rewrite"""

    prog = \
"""builtin.module() {
%0 : !i32 = arith.constant() ["value" = 42 : !i32]
%1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 43 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    class RewriteConst(RewritePattern):

        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            if isinstance(op, Constant):
                new_constant = Constant.from_int_and_width(43, i32)
                rewriter.replace_matched_op([new_constant])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(RewriteConst(), apply_recursively=False))


def test_non_recursive_rewrite_reversed():
    """Test a simple non-recursive rewrite with reverse walk order."""

    prog = \
    """builtin.module() {
%0 : !i32 = arith.constant() ["value" = 42 : !i32]
%1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 43 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    class RewriteConst(RewritePattern):

        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            if isinstance(op, Constant):
                new_constant = Constant.from_int_and_width(43, i32)
                rewriter.replace_matched_op([new_constant])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(RewriteConst(),
                             apply_recursively=False,
                             walk_reverse=True))


def test_op_type_rewrite_pattern_method_decorator():
    """Test op_type_rewrite_pattern decorator on methods."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 43 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    class RewriteConst(RewritePattern):

        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: Constant, rewriter: PatternRewriter):
            rewriter.replace_matched_op(Constant.from_int_and_width(43, i32))

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(RewriteConst(), apply_recursively=False))


def test_op_type_rewrite_pattern_static_decorator():
    """Test op_type_rewrite_pattern decorator on static functions."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 43 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: Constant, rewriter: PatternRewriter):
        rewriter.replace_matched_op(Constant.from_int_and_width(43, i32))

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_recursive_rewriter():
    """Test recursive walks on operations created by rewrites."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 1 : !i64]
  %1 : !i32 = arith.constant() ["value" = 1 : !i64]
  %2 : !i32 = arith.addi(%0 : !i32, %1 : !i32)
  %3 : !i32 = arith.constant() ["value" = 1 : !i64]
  %4 : !i32 = arith.addi(%2 : !i32, %3 : !i32)
  %5 : !i32 = arith.constant() ["value" = 1 : !i64]
  %6 : !i32 = arith.addi(%4 : !i32, %5 : !i32)
  %7 : !i32 = arith.constant() ["value" = 1 : !i64]
  %8 : !i32 = arith.addi(%6 : !i32, %7 : !i32)
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: Constant, rewriter: PatternRewriter):
        val = op.value.value.data
        if val == 0 or val == 1:
            return None
        constant_op = Constant.from_attr(
            IntegerAttr.from_int_and_width(val - 1, 64), i32)
        constant_one = Constant.from_attr(
            IntegerAttr.from_int_and_width(1, 64), i32)
        add_op = Addi.get(constant_op, constant_one)
        rewriter.replace_matched_op([constant_op, constant_one, add_op])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=True))


def test_recursive_rewriter_reversed():
    """Test recursive walks on operations created by rewrites, in reverse walk order."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 1 : !i64]
  %1 : !i32 = arith.constant() ["value" = 1 : !i64]
  %2 : !i32 = arith.addi(%0 : !i32, %1 : !i32)
  %3 : !i32 = arith.constant() ["value" = 1 : !i64]
  %4 : !i32 = arith.addi(%2 : !i32, %3 : !i32)
  %5 : !i32 = arith.constant() ["value" = 1 : !i64]
  %6 : !i32 = arith.addi(%4 : !i32, %5 : !i32)
  %7 : !i32 = arith.constant() ["value" = 1 : !i64]
  %8 : !i32 = arith.addi(%6 : !i32, %7 : !i32)
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: Constant, rewriter: PatternRewriter):
        val = op.value.value.data
        if val == 0 or val == 1:
            return None
        constant_op = Constant.from_attr(
            IntegerAttr.from_int_and_width(val - 1, 64), i32)
        constant_one = Constant.from_attr(
            IntegerAttr.from_int_and_width(1, 64), i32)
        add_op = Addi.get(constant_op, constant_one)
        rewriter.replace_matched_op([constant_op, constant_one, add_op])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=True,
                             walk_reverse=True))


def test_greedy_rewrite_pattern_applier():
    """Test GreedyRewritePatternApplier."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 43 : !i32]
  %1 : !i32 = arith.muli(%0 : !i32, %0 : !i32)
}"""

    @op_type_rewrite_pattern
    def constant_rewrite(op: Constant, rewriter: PatternRewriter):
        rewriter.replace_matched_op([Constant.from_int_and_width(43, i32)])

    @op_type_rewrite_pattern
    def addi_rewrite(op: Addi, rewriter: PatternRewriter):
        rewriter.replace_matched_op([Muli.get(op.lhs, op.rhs)])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(GreedyRewritePatternApplier([
            AnonymousRewritePattern(constant_rewrite),
            AnonymousRewritePattern(addi_rewrite)
        ]),
                             apply_recursively=False))


def test_insert_op_before_matched_op():
    """Test rewrites where operations are inserted before the matched operation."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  %1 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(cst: Constant, rewriter: PatternRewriter):
        new_cst = Constant.from_int_and_width(42, i32)
        rewriter.insert_op_before_matched_op(new_cst)

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_insert_op_at_pos():
    """Test rewrites where operations are inserted with a given position."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  %1 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(mod: ModuleOp, rewriter: PatternRewriter):
        new_cst = Constant.from_int_and_width(42, i32)
        rewriter.insert_op_at_pos(new_cst, mod.regions[0].blocks[0], 0)

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_insert_op_before():
    """Test rewrites where operations are inserted before a given operation."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
  %1 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(mod: ModuleOp, rewriter: PatternRewriter):
        new_cst = Constant.from_int_and_width(42, i32)
        rewriter.insert_op_before(new_cst, mod.ops[0])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_insert_op_after():
    """Test rewrites where operations are inserted after a given operation."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
  %1 : !i32 = arith.constant() ["value" = 42 : !i32]
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(mod: ModuleOp, rewriter: PatternRewriter):
        new_cst = Constant.from_int_and_width(42, i32)
        rewriter.insert_op_after(new_cst, mod.ops[0])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_insert_op_after_matched_op():
    """Test rewrites where operations are inserted after a given operation."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
  %1 : !i32 = arith.constant() ["value" = 42 : !i32]
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(cst: Constant, rewriter: PatternRewriter):
        new_cst = Constant.from_int_and_width(42, i32)
        rewriter.insert_op_after_matched_op(new_cst)

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_insert_op_after_matched_op_reversed():
    """Test rewrites where operations are inserted after a given operation."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
  %1 : !i32 = arith.constant() ["value" = 42 : !i32]
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(cst: Constant, rewriter: PatternRewriter):
        new_cst = Constant.from_int_and_width(42, i32)
        rewriter.insert_op_after_matched_op(new_cst)

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False,
                             walk_reverse=True))


def test_operation_deletion():
    """Test rewrites where SSA values are deleted."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    expected = \
"""builtin.module() {}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: Constant, rewriter: PatternRewriter):
        rewriter.erase_matched_op()

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite)))


def test_operation_deletion_reversed():
    """
    Test rewrites where SSA values are deleted.
    They have to be deleted in order for the rewrite to not fail.
    """

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    expected = \
"""builtin.module() {}"""

    def match_and_rewrite(op: Operation, rewriter: PatternRewriter):
        if not isinstance(op, ModuleOp):
            rewriter.erase_matched_op()

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             walk_reverse=True))


def test_operation_deletion_failure():
    """Test rewrites where SSA values are deleted with still uses."""

    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_dialect(Arith)

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
  %1 : !i32 = arith.addi(%0 : !i32, %0 : !i32)
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: Constant, rewriter: PatternRewriter):
        rewriter.erase_matched_op()

    parser = Parser(ctx, prog)
    module = parser.parse_op()
    walker = PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite))

    # Check that the rewrite fails
    try:
        walker.rewrite_module(module)
        assert False
    except Exception:
        pass


def test_delete_inner_op():
    """Test rewrites where an operation inside a region of the matched op is deleted."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    expected = \
"""builtin.module() {}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: ModuleOp, rewriter: PatternRewriter):
        rewriter.erase_op(op.ops[0])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite)))


def test_replace_inner_op():
    """Test rewrites where an operation inside a region of the matched op is deleted."""

    prog = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 5 : !i32]
}"""

    expected = \
"""builtin.module() {
  %0 : !i32 = arith.constant() ["value" = 42 : !i32]
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: ModuleOp, rewriter: PatternRewriter):
        rewriter.replace_op(op.ops[0], [Constant.from_int_and_width(42, i32)])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite)))


def test_block_argument_type_change():
    """Test the modification of a block argument type."""

    prog = \
    """builtin.module() {
%0 : !i1 = arith.constant() ["value" = true]
scf.if(%0 : !i1) {
^0(%1 : !i32):
  %2 : !i32 = arith.addi(%1 : !i32, %1 : !i32)
} {}
}"""

    expected = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  scf.if(%0 : !i1) {
  ^0(%1 : !i64):
    %2 : !i32 = arith.addi(%1 : !i64, %1 : !i64)
  } {}
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: If, rewriter: PatternRewriter):
        rewriter.modify_block_argument_type(op.true_region.blocks[0].args[0],
                                            i64)

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_block_argument_erasure():
    """Test the erasure of a block argument."""

    prog = \
    """builtin.module() {
%0 : !i1 = arith.constant() ["value" = true]
scf.if(%0 : !i1) {
^0(%1 : !i32):
} {}
}"""

    expected = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  scf.if(%0 : !i1) {} {}
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: If, rewriter: PatternRewriter):
        rewriter.erase_block_argument(op.true_region.blocks[0].args[0])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_block_argument_insertion():
    """Test the insertion of a block argument."""

    prog = \
    """builtin.module() {
%0 : !i1 = arith.constant() ["value" = true]
scf.if(%0 : !i1) {} {}
}"""

    expected = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  scf.if(%0 : !i1) {
  ^0(%1 : !i32):
  } {}
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: If, rewriter: PatternRewriter):
        rewriter.insert_block_argument(op.true_region.blocks[0], 0, i32)

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_inline_block_at_pos():
    """Test the inlining of a block at a certain position."""

    prog = \
    """builtin.module() {
%0 : !i1 = arith.constant() ["value" = true]
scf.if(%0 : !i1) {
  scf.if(%0 : !i1) {
    %1 : !i32 = arith.constant() ["value" = 2 : !i32]
  } {}
} {}
}"""

    expected = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  scf.if(%0 : !i1) {
    %1 : !i32 = arith.constant() ["value" = 2 : !i32]
    scf.if(%0 : !i1) {} {}
  } {}
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: If, rewriter: PatternRewriter):
        if len(op.true_region.blocks[0].ops) > 0 and isinstance(
                op.true_region.blocks[0].ops[0], If):
            inner_if_block = op.true_region.blocks[0].ops[
                0].true_region.blocks[0]
            rewriter.inline_block_at_pos(inner_if_block,
                                         op.true_region.blocks[0], 0)

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite)))


def test_inline_block_before_matched_op():
    """Test the inlining of a block before the matched operation."""

    prog = \
    """builtin.module() {
%0 : !i1 = arith.constant() ["value" = true]
scf.if(%0 : !i1) {
  %1 : !i32 = arith.constant() ["value" = 2 : !i32]
} {}
}"""

    expected = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  %1 : !i32 = arith.constant() ["value" = 2 : !i32]
  scf.if(%0 : !i1) {} {}
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: If, rewriter: PatternRewriter):
        rewriter.inline_block_before_matched_op(op.true_region.blocks[0])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_inline_block_before():
    """Test the inlining of a block before an operation."""

    prog = \
"""builtin.module() {
%0 : !i1 = arith.constant() ["value" = true]
scf.if(%0 : !i1) {
  scf.if(%0 : !i1) {
    %1 : !i32 = arith.constant() ["value" = 2 : !i32]
  } {}
} {}
}"""

    expected = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  scf.if(%0 : !i1) {
    %1 : !i32 = arith.constant() ["value" = 2 : !i32]
    scf.if(%0 : !i1) {} {}
  } {}
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: If, rewriter: PatternRewriter):
        if len(op.true_region.blocks[0].ops) > 0 and isinstance(
                op.true_region.blocks[0].ops[0], If):
            inner_if_block = op.true_region.blocks[0].ops[
                0].true_region.blocks[0]
            rewriter.inline_block_before(inner_if_block,
                                         op.true_region.blocks[0].ops[0])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_inline_block_at_before_when_op_is_matched_op():
    """Test the inlining of a block before an operation, being the matched one."""

    prog = \
    """builtin.module() {
%0 : !i1 = arith.constant() ["value" = true]
scf.if(%0 : !i1) {
  %1 : !i32 = arith.constant() ["value" = 2 : !i32]
} {}
}"""

    expected = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  %1 : !i32 = arith.constant() ["value" = 2 : !i32]
  scf.if(%0 : !i1) {} {}
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: If, rewriter: PatternRewriter):
        rewriter.inline_block_before(op.true_region.blocks[0], op)

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_inline_block_after():
    """Test the inlining of a block after an operation."""

    prog = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  scf.if(%0 : !i1) {
    scf.if(%0 : !i1) {
      %1 : !i32 = arith.constant() ["value" = 2 : !i32]
    } {}
  } {}
}"""

    expected = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  scf.if(%0 : !i1) {
    scf.if(%0 : !i1) {} {}
    %1 : !i32 = arith.constant() ["value" = 2 : !i32]
  } {}
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: If, rewriter: PatternRewriter):
        if len(op.true_region.blocks[0].ops) > 0 and isinstance(
                op.true_region.blocks[0].ops[0], If):
            inner_if_block = op.true_region.blocks[0].ops[
                0].true_region.blocks[0]
            rewriter.inline_block_after(inner_if_block,
                                        op.true_region.blocks[0].ops[0])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))


def test_move_region_contents_to_new_regions():
    """Test moving a region outside of a region."""

    prog = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  scf.if(%0 : !i1) {
    %1 : !i32 = arith.constant() ["value" = 2 : !i32]
  }
}"""

    expected = \
"""builtin.module() {
  %0 : !i1 = arith.constant() ["value" = true]
  scf.if(%0 : !i1) {}
  scf.if(%0 : !i1) {
    %1 : !i32 = arith.constant() ["value" = 2 : !i32]
  } {}
}"""

    @op_type_rewrite_pattern
    def match_and_rewrite(op: ModuleOp, rewriter: PatternRewriter):
        new_region = rewriter.move_region_contents_to_new_regions(
            op.ops[1].regions[0])
        new_if = If.get(op.ops[1].cond, [], new_region,
                        Region.from_operation_list([]))
        rewriter.insert_op_after(new_if, op.ops[1])

    rewrite_and_compare(
        prog, expected,
        PatternRewriteWalker(AnonymousRewritePattern(match_and_rewrite),
                             apply_recursively=False))



from io import StringIO

import pytest

from xdsl.printer import Printer
from xdsl.ir import MLContext, Attribute
from xdsl.parser import XDSLParser
from xdsl.dialects.builtin import IntAttr, DictionaryAttr, StringAttr, ArrayAttr, Builtin


@pytest.mark.parametrize("input,expected", [("0, 1, 1", [0, 1, 1]),
                                            ("1, 0, 1", [1, 0, 1]),
                                            ("1, 1, 0", [1, 1, 0])])
def test_int_list_parser(input: str, expected: list[int]):
    ctx = MLContext()
    parser = XDSLParser(ctx, input)

    int_list = parser.parse_list_of(parser.try_parse_integer_literal, '')
    assert [int(span.text) for span in int_list] == expected


@pytest.mark.parametrize('data', [
    dict(a=IntAttr.from_int(1), b=IntAttr.from_int(2), c=IntAttr.from_int(3)),
    dict(a=StringAttr.from_str('hello'),
         b=IntAttr.from_int(2),
         c=ArrayAttr.from_list(
             [IntAttr.from_int(2),
              StringAttr.from_str('world')])),
    dict(),
])
def test_dictionary_attr(data: dict[str, Attribute]):
    attr = DictionaryAttr.from_dict(data)

    with StringIO() as io:
        Printer(io).print(attr)
        text = io.getvalue()

    ctx = MLContext()
    ctx.register_dialect(Builtin)

    attr = XDSLParser(ctx, text).parse_attribute()

    assert attr.data == data



from __future__ import annotations
from typing import Annotated
from xdsl.ir import Attribute, Dialect, OpResult, SSAValue
from xdsl.irdl import Operand, irdl_op_definition, OpAttr, AnyAttr, Operation
from xdsl.dialects.builtin import StringAttr, FlatSymbolRefAttr


@irdl_op_definition
class Declare(Operation):
    name: str = "symref.declare"
    sym_name: OpAttr[StringAttr]

    @staticmethod
    def get(sym_name: str | StringAttr) -> Declare:
        return Declare.build(attributes={"sym_name": sym_name})


@irdl_op_definition
class Fetch(Operation):
    name: str = "symref.fetch"
    value: Annotated[OpResult, AnyAttr()]
    symbol: OpAttr[FlatSymbolRefAttr]

    @staticmethod
    def get(symbol: str | FlatSymbolRefAttr, result_type: Attribute) -> Fetch:
        return Fetch.build(attributes={"symbol": symbol},
                           result_types=[result_type])


@irdl_op_definition
class Update(Operation):
    name: str = "symref.update"
    value: Annotated[Operand, AnyAttr()]
    symbol: OpAttr[FlatSymbolRefAttr]

    @staticmethod
    def get(symbol: str | FlatSymbolRefAttr,
            value: Operation | SSAValue) -> Update:
        return Update.build(operands=[value], attributes={"symbol": symbol})


Symref = Dialect([Declare, Fetch, Update], [])



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from frozenlist import FrozenList
from io import StringIO
from typing import (TYPE_CHECKING, Any, Callable, Generic, Optional, Protocol,
                    Sequence, TypeVar, cast, Iterator, Union, ClassVar)
import sys

# Used for cyclic dependencies in type hints
if TYPE_CHECKING:
    from xdsl.parser import Parser, BaseParser
    from xdsl.printer import Printer
    from xdsl.irdl import OpDef, ParamAttrDef

OpT = TypeVar('OpT', bound='Operation')


@dataclass
class Dialect:
    """Contains the operations and attributes of a specific dialect"""
    _operations: list[type[Operation]] = field(default_factory=list,
                                               init=True,
                                               repr=True)
    _attributes: list[type[Attribute]] = field(default_factory=list,
                                               init=True,
                                               repr=True)

    @property
    def operations(self) -> Iterator[type[Operation]]:
        return iter(self._operations)

    @property
    def attributes(self) -> Iterator[type[Attribute]]:
        return iter(self._attributes)

    def __call__(self, ctx: MLContext) -> None:
        print(
            "Calling a dialect in order to register it is deprecated and will soon be removed.",
            file=sys.stderr)
        # TODO; Remove this function in a future release.
        assert isinstance(ctx, MLContext)
        ctx.register_dialect(self)


@dataclass
class MLContext:
    """Contains structures for operations/attributes registration."""
    _registeredOps: dict[str, type[Operation]] = field(default_factory=dict)
    _registeredAttrs: dict[str, type[Attribute]] = field(default_factory=dict)
    registered_unregistered_ops: dict[str, type[Operation]] = field(
        default_factory=dict)

    def register_dialect(self, dialect: Dialect):
        """Register a dialect. Operation and Attribute names should be unique"""
        for op in dialect.operations:
            self.register_op(op)

        for attr in dialect.attributes:
            self.register_attr(attr)

    def register_op(self, op: type[Operation]) -> None:
        """Register an operation definition. Operation names should be unique."""
        if op.name in self._registeredOps:
            raise Exception(f"Operation {op.name} has already been registered")
        self._registeredOps[op.name] = op

    def register_attr(self, attr: type[Attribute]) -> None:
        """Register an attribute definition. Attribute names should be unique."""
        if attr.name in self._registeredAttrs:
            raise Exception(
                f"Attribute {attr.name} has already been registered")
        self._registeredAttrs[attr.name] = attr

    def get_optional_op(self, name: str) -> type[Operation] | None:
        """Get an operation class from its name if it exists."""
        if name not in self._registeredOps:
            return None
        return self._registeredOps[name]

    def get_op(self, name: str) -> type[Operation]:
        """Get an operation class from its name."""
        if op_type := self.get_optional_op(name):
            return op_type
        raise Exception(f"Operation {name} is not registered")

    def get_optional_attr(self, name: str) -> type[Attribute] | None:
        """Get an attribute class from its name if it exists."""
        if name not in self._registeredAttrs:
            return None
        return self._registeredAttrs[name]

    def get_attr(self, name: str) -> type[Attribute]:
        """Get an attribute class from its name."""
        if attr_type := self.get_optional_attr(name):
            return attr_type
        raise Exception(f"Attribute {name} is not registered")


@dataclass(frozen=True)
class Use:
    """The use of a SSA value."""

    operation: Operation
    """The operation using the value."""

    index: int
    """The index of the operand using the value in the operation."""


@dataclass
class SSAValue(ABC):
    """A reference to an SSA variable.
    An SSA variable is either an operation result, or a basic block argument."""

    typ: Attribute
    """Each SSA variable is associated to a type."""

    uses: set[Use] = field(init=False, default_factory=set, repr=False)
    """All uses of the value."""

    _name: str | None = field(init=False, default=None)

    _name_regex: ClassVar[re.Pattern] = re.compile(
        r'[A-Za-z0-9._$-]*[A-Za-z._$-]')

    @property
    def name(self) -> str | None:
        return self._name

    @name.setter
    def name(self, name: str):
        if self._name_regex.fullmatch(name):
            self._name = name

    @staticmethod
    def get(arg: SSAValue | Operation) -> SSAValue:
        "Get a new SSAValue from either a SSAValue, or an operation with a single result."
        if isinstance(arg, SSAValue):
            return arg
        if isinstance(arg, Operation):
            if len(arg.results) == 1:
                return arg.results[0]
            raise ValueError(
                "SSAValue.build: expected operation with a single result.")
        raise TypeError(
            f"Expected SSAValue or Operation for SSAValue.get, but got {arg}")

    def add_use(self, use: Use):
        """Add a new use of the value."""
        self.uses.add(use)

    def remove_use(self, use: Use):
        """Remove a use of the value."""
        assert use in self.uses, "use to be removed was not in use list"
        self.uses.remove(use)

    def replace_by(self, value: SSAValue) -> None:
        """Replace the value by another value in all its uses."""
        for use in self.uses.copy():
            use.operation.replace_operand(use.index, value)
        assert len(self.uses) == 0, "unexpected error in xdsl"

    def erase(self, safe_erase: bool = True) -> None:
        """
        Erase the value.
        If safe_erase is True, then check that no operations use the value anymore.
        If safe_erase is False, then replace its uses by an ErasedSSAValue.
        """
        if safe_erase and len(self.uses) != 0:
            raise Exception(
                "Attempting to delete SSA value that still has uses of operation:\n"
                f"{self.op}")
        self.replace_by(ErasedSSAValue(self.typ, self))


@dataclass
class OpResult(SSAValue):
    """A reference to an SSA variable defined by an operation result."""

    op: Operation
    """The operation defining the variable."""

    result_index: int
    """The index of the result in the defining operation."""

    def __repr__(self) -> str:
        return f"OpResult(typ={repr(self.typ)}, num_uses={repr(len(self.uses))}, " + \
               f"op_name={repr(self.op.name)}, " + \
               f"result_index={repr(self.result_index)}, name={repr(self.name)})"

    def __eq__(self, other: object) -> bool:
        return self is other

    # This might be problematic, as the superclass is not hashable ...
    def __hash__(self) -> int:  # type: ignore
        return id(self)


@dataclass
class BlockArgument(SSAValue):
    """A reference to an SSA variable defined by a basic block argument."""

    block: Block
    """The block defining the variable."""

    index: int
    """The index of the variable in the block arguments."""

    def __repr__(self) -> str:
        if isinstance(self.block, Block):
            block_repr = f"Block(num_arguments={len(self.block.args)}, " + \
                         f"num_blocks={len(self.block.ops)} ops)"
        else:
            block_repr = repr(self.block)
        return f"OpResult(typ={repr(self.typ)}, num_uses={repr(len(self.uses))}" + \
               f", block={block_repr}, index={repr(self.index)}"

    def __eq__(self, other: object) -> bool:
        return self is other

    def __hash__(self) -> int:  # type: ignore
        return id(self)


@dataclass
class ErasedSSAValue(SSAValue):
    """
    An erased SSA variable.
    This is used during transformations when a SSA variable is destroyed but still used.
    """

    old_value: SSAValue

    def __hash__(self) -> int:  # type: ignore
        return hash(id(self))


@dataclass
class MLIRType:
    """
    A class representing an MLIR type.
    This class should only be inherited by classes inheriting Attribute.
    This class is only used for printing attributes in the MLIR format,
    inheriting this class prefix the attribute by `!` instead of `#`.
    """

    def __post_init__(self):
        if not isinstance(self, Attribute):
            raise TypeError(
                "MLIRType should only be inherited by classes inheriting Attribute"
            )


A = TypeVar('A', bound='Attribute')


class Attribute(ABC):
    """
    A compile-time value.
    Attributes are used to represent SSA variable types, and can be attached
    on operations to give extra information.
    """
    name: str = field(default="", init=False)
    """The attribute name should be a static field in the attribute classes."""

    @classmethod
    def build(cls: type[A], *args: Any) -> A:
        """Create a new attribute using one of the builder defined in IRDL."""
        assert False

    def __post_init__(self):
        self.verify()

    def verify(self) -> None:
        """
        Check that the attribute parameters satisfy the expected invariants.
        Raise an exception otherwise.
        """
        pass

    def __str__(self) -> str:
        from xdsl.printer import Printer
        res = StringIO()
        printer = Printer(stream=res)
        printer.print_attribute(self)
        return res.getvalue()


DataElement = TypeVar("DataElement")


@dataclass(frozen=True)
class Data(Generic[DataElement], Attribute, ABC):
    """An attribute represented by a Python structure."""
    data: DataElement

    @staticmethod
    @abstractmethod
    def parse_parameter(parser: BaseParser) -> DataElement:
        """Parse the attribute parameter."""

    @staticmethod
    @abstractmethod
    def print_parameter(data: DataElement, printer: Printer) -> None:
        """Print the attribute parameter."""


@dataclass(frozen=True)
class ParametrizedAttribute(Attribute):
    """An attribute parametrized by other attributes."""
    parameters: list[Attribute] = field(default_factory=list)

    @staticmethod
    def parse_parameters(parser: BaseParser) -> list[Attribute]:
        """Parse the attribute parameters."""
        return parser.parse_paramattr_parameters()

    def print_parameters(self, printer: Printer) -> None:
        """Print the attribute parameters."""
        printer.print_paramattr_parameters(self.parameters)

    @classmethod
    @property
    def irdl_definition(cls) -> ParamAttrDef:
        """Get the IRDL attribute definition."""
        ...


@dataclass
class IRNode(ABC):

    parent: IRNode | None

    def is_ancestor(self, op: IRNode) -> bool:
        "Returns true if the IRNode is an ancestor of another IRNode."
        if op is self:
            return True
        if op.parent is None:
            return False
        return self.is_ancestor(op.parent)

    def get_toplevel_object(self) -> IRNode:
        """Get the operation, block, or region ancestor that has no parents."""
        if self.parent is None:
            return self
        return self.parent.get_toplevel_object()

    def is_structurally_equivalent(
            self,
            other: IRNode,
            context: Optional[dict[IRNode, IRNode]] = None) -> bool:
        """Check if two IR nodes are structurally equivalent."""
        ...


@dataclass
class Operation(IRNode):
    """A generic operation. Operation definitions inherit this class."""

    name: str = field(default="", init=False)
    """The operation name. Should be a static member of the class"""

    _operands: FrozenList[SSAValue] = field(default_factory=FrozenList)
    """The operation operands."""

    results: list[OpResult] = field(default_factory=list)
    """The results created by the operation."""

    successors: list[Block] = field(default_factory=list)
    """
    The basic blocks that the operation may give control to.
    This list should be empty for non-terminator operations.
    """

    attributes: dict[str, Attribute] = field(default_factory=dict)
    """The attributes attached to the operation."""

    regions: list[Region] = field(default_factory=list)
    """Regions arguments of the operation."""

    parent: Block | None = field(default=None, repr=False)
    """The block containing this operation."""

    def parent_op(self) -> Operation | None:
        if p := self.parent_region():
            return p.parent
        return None

    def parent_region(self) -> Region | None:
        if p := self.parent_block():
            return p.parent
        return None

    def parent_block(self) -> Block | None:
        return self.parent

    @property
    def operands(self) -> FrozenList[SSAValue]:
        return self._operands

    @operands.setter
    def operands(self, new: list[SSAValue] | FrozenList[SSAValue]):
        if isinstance(new, list):
            new = FrozenList(new)
        for idx, operand in enumerate(self._operands):
            operand.remove_use(Use(self, idx))
        for idx, operand in enumerate(new):
            operand.add_use(Use(self, idx))
        self._operands = new
        self._operands.freeze()

    def __post_init__(self):
        assert (self.name != "")
        assert (isinstance(self.name, str))

    @staticmethod
    def with_result_types(
            op: Any,
            operands: Sequence[SSAValue] | None = None,
            result_types: Sequence[Attribute] | None = None,
            attributes: dict[str, Attribute] | None = None,
            successors: Sequence[Block] | None = None,
            regions: Sequence[Region] | None = None) -> Operation:

        operation = op()
        if operands is not None:
            operation.operands = operands
        if result_types:
            operation.results = [
                OpResult(typ, operation, idx)
                for (idx, typ) in enumerate(result_types)
            ]
        if attributes:
            operation.attributes = attributes
        if successors:
            operation.successors = successors
        if regions:
            for region in regions:
                operation.add_region(region)
        return operation

    @classmethod
    def create(cls: type[OpT],
               operands: Sequence[SSAValue] | None = None,
               result_types: Sequence[Attribute] | None = None,
               attributes: dict[str, Attribute] | None = None,
               successors: Sequence[Block] | None = None,
               regions: Sequence[Region] | None = None) -> OpT:
        op = Operation.with_result_types(cls, operands, result_types,
                                         attributes, successors, regions)
        return cast(OpT, op)

    @classmethod
    def build(cls: type[OpT],
              operands: list[Any] | None = None,
              result_types: list[Any] | None = None,
              attributes: dict[str, Any] | None = None,
              successors: list[Any] | None = None,
              regions: list[Any] | None = None) -> OpT:
        """Create a new operation using builders."""
        ...

    def replace_operand(self, operand_idx: int, new_operand: SSAValue) -> None:
        """Replace an operand with another operand."""
        self.operands = list(self._operands[:operand_idx]) + [
            new_operand
        ] + list(self._operands[operand_idx + 1:])

    def add_region(self, region: Region) -> None:
        """Add an unattached region to the operation."""
        if region.parent:
            raise Exception(
                "Cannot add region that is already attached on an operation.")
        self.regions.append(region)
        region.parent = self

    def drop_all_references(self) -> None:
        """
        Drop all references to other operations.
        This function is called prior to deleting an operation.
        """
        self.parent = None
        for idx, operand in enumerate(self.operands):
            operand.remove_use(Use(self, idx))
        for region in self.regions:
            region.drop_all_references()

    def walk(self, fun: Callable[[Operation], None]) -> None:
        """
        Call a function on all operations contained in the operation (including this one)
        """
        fun(self)
        for region in self.regions:
            region.walk(fun)

    def verify(self, verify_nested_ops: bool = True) -> None:
        for operand in self.operands:
            if isinstance(operand, ErasedSSAValue):
                raise Exception("Erased SSA value is used by the operation")
        self.verify_()
        if verify_nested_ops:
            for region in self.regions:
                region.verify()

    def verify_(self) -> None:
        pass

    _OperationType = TypeVar('_OperationType', bound='Operation')

    @classmethod
    def parse(cls: type[_OperationType], result_types: list[Attribute],
              parser: BaseParser) -> _OperationType:
        return parser.parse_op_with_default_format(cls, result_types)

    def print(self, printer: Printer):
        return printer.print_op_with_default_format(self)

    def clone_without_regions(
            self: OpT,
            value_mapper: dict[SSAValue, SSAValue] | None = None,
            block_mapper: dict[Block, Block] | None = None) -> OpT:
        """Clone an operation, with empty regions instead."""
        if value_mapper is None:
            value_mapper = {}
        if block_mapper is None:
            block_mapper = {}
        operands = [
            (value_mapper[operand] if operand in value_mapper else operand)
            for operand in self.operands
        ]
        result_types = [res.typ for res in self.results]
        attributes = self.attributes.copy()
        successors = [(block_mapper[successor]
                       if successor in block_mapper else successor)
                      for successor in self.successors]
        regions = [Region() for _ in self.regions]
        cloned_op = self.create(operands=operands,
                                result_types=result_types,
                                attributes=attributes,
                                successors=successors,
                                regions=regions)
        for idx, result in enumerate(cloned_op.results):
            value_mapper[self.results[idx]] = result
        return cloned_op

    def clone(self: OpT,
              value_mapper: dict[SSAValue, SSAValue] | None = None,
              block_mapper: dict[Block, Block] | None = None) -> OpT:
        """Clone an operation with all its regions and operations in them."""
        if value_mapper is None:
            value_mapper = {}
        if block_mapper is None:
            block_mapper = {}
        op = self.clone_without_regions(value_mapper, block_mapper)
        for idx, region in enumerate(self.regions):
            region.clone_into(op.regions[idx], 0, value_mapper, block_mapper)
        return op

    def erase(self,
              safe_erase: bool = True,
              drop_references: bool = True) -> None:
        """
        Erase the operation, and remove all its references to other operations.
        If safe_erase is specified, check that the operation results are not used.
        """
        assert self.parent is None, "Operation with parents should first be detached " + \
                                    "before erasure."
        if drop_references:
            self.drop_all_references()
        for result in self.results:
            result.erase(safe_erase=safe_erase)

    def detach(self):
        """Detach the operation from its parent block."""
        if self.parent is None:
            raise Exception("Cannot detach a toplevel operation.")
        self.parent.detach_op(self)

    def is_structurally_equivalent(
            self,
            other: IRNode,
            context: Optional[dict[IRNode, IRNode]] = None) -> bool:
        """
        Check if two operations are structurally equivalent.
        The context is a mapping of IR nodes to IR nodes that are already known to be equivalent.
        This enables checking whether the use dependencies and successors are equivalent.
        """
        if context is None:
            context = {}
        if not isinstance(other, Operation):
            return False
        if self.name != other.name:
            return False
        if len(self.operands) != len(other.operands) or \
           len(self.results) != len(other.results) or \
           len(self.regions) != len(other.regions) or \
           len(self.successors) != len(other.successors) or \
            self.attributes != other.attributes:
            return False
        if self.parent and other.parent and context.get(
                self.parent) != other.parent:
            return False
        if not all(
                context.get(operand) == other_operand for operand,
                other_operand in zip(self.operands, other.operands)):
            return False
        if not all(
                context.get(successor) == other_successor for successor,
                other_successor in zip(self.successors, other.successors)):
            return False
        if not all(
                region.is_structurally_equivalent(other_region, context)
                for region, other_region in zip(self.regions, other.regions)):
            return False
        # Add results of this operation to the context
        for result, other_result in zip(self.results, other.results):
            context[result] = other_result

        return True

    def __eq__(self, other: object) -> bool:
        return self is other

    def __hash__(self) -> int:
        return id(self)

    @classmethod
    @property
    def irdl_definition(cls) -> OpDef:
        """Get the IRDL operation definition."""
        ...


@dataclass()
class Block(IRNode):
    """A sequence of operations"""

    declared_at: 'Span' | None = None

    _args: FrozenList[BlockArgument] = field(default_factory=FrozenList,
                                             init=False)
    """The basic block arguments."""

    ops: list[Operation] = field(default_factory=list, init=False)
    """Ordered operations contained in the block."""

    parent: Region | None = field(default=None, init=False, repr=False)
    """Parent region containing the block."""

    def parent_op(self) -> Operation | None:
        return self.parent.parent if self.parent else None

    def parent_region(self) -> Region | None:
        return self.parent

    def parent_block(self) -> Block | None:
        return self.parent.parent.parent if self.parent and self.parent.parent else None

    def __repr__(self) -> str:
        return f"Block(_args={repr(self._args)}, num_ops={len(self.ops)})"

    @property
    def args(self) -> FrozenList[BlockArgument]:
        """Returns the block arguments."""
        return self._args

    @staticmethod
    def from_arg_types(arg_types: list[Attribute]) -> Block:
        b = Block()
        b._args = FrozenList([
            BlockArgument(typ, b, index) for index, typ in enumerate(arg_types)
        ])
        b._args.freeze()
        return b

    @staticmethod
    def from_ops(ops: list[Operation],
                 arg_types: list[Attribute] | None = None):
        b = Block()
        if arg_types:
            b._args = FrozenList([
                BlockArgument(typ, b, index)
                for index, typ in enumerate(arg_types)
            ])
            b._args.freeze()
        b.add_ops(ops)
        return b

    class BlockCallback(Protocol):

        def __call__(self, *args: BlockArgument) -> list[Operation]:
            ...

    @staticmethod
    def from_callable(block_arg_types: list[Attribute], f: BlockCallback):
        b = Block.from_arg_types(block_arg_types)
        b.add_ops(f(*b.args))
        return b

    def insert_arg(self, typ: Attribute, index: int) -> BlockArgument:
        """
        Insert a new argument with a given type to the arguments list at a specific index.
        Returns the new argument.
        """
        if index < 0 or index > len(self._args):
            raise Exception("Unexpected index")
        new_arg = BlockArgument(typ, self, index)
        for arg in self._args[index:]:
            arg.index += 1
        self._args = FrozenList(
            list(self._args[:index]) + [new_arg] + list(self._args[index:]))
        self._args.freeze()
        return new_arg

    def erase_arg(self, arg: BlockArgument, safe_erase: bool = True) -> None:
        """
        Erase a block argument.
        If safe_erase is True, check that the block argument is not used.
        If safe_erase is False, replace the block argument uses with an ErasedSSAVAlue.
        """
        if arg.block is not self:
            raise Exception(
                "Attempting to delete an argument of the wrong block")
        for block_arg in self._args[arg.index + 1:]:
            block_arg.index -= 1
        self._args = FrozenList(
            list(self._args[:arg.index]) + list(self._args[arg.index + 1:]))
        arg.erase(safe_erase=safe_erase)

    def _attach_op(self, operation: Operation) -> None:
        """Attach an operation to the block, and check that it has no parents."""
        if operation.parent:
            raise ValueError(
                "Can't add to a block an operation already attached to a block."
            )
        if operation.is_ancestor(self):
            raise ValueError(
                "Can't add an operation to a block contained in the operation."
            )
        operation.parent = self

    def add_op(self, operation: Operation) -> None:
        """
        Add an operation at the end of the block.
        The operation should not be attached to another block already.
        """
        self._attach_op(operation)
        self.ops.append(operation)

    def add_ops(self, ops: list[Operation]) -> None:
        """
        Add operations at the end of the block.
        The operations should not be attached to another block.
        """
        for op in ops:
            self.add_op(op)

    def insert_op(self,
                  ops: Operation | list[Operation],
                  index: int,
                  name: str | None = None) -> None:
        """
        Insert one or multiple operations at a given index in the block.
        The operations should not be attached to another block.
        """
        if index < 0 or index > len(self.ops):
            raise ValueError(
                f"Can't insert operation in index {index} in a block with "
                f"{len(self.ops)} operations.")
        if not isinstance(ops, list):
            ops = [ops]
        if name:
            for curr_op in ops:
                for res in curr_op.results:
                    res.name = name
        for op in ops:
            self._attach_op(op)
        self.ops = self.ops[:index] + ops + self.ops[index:]

    def get_operation_index(self, op: Operation) -> int:
        """Get the operation position in a block."""
        if op.parent is not self:
            raise Exception("Operation is not a children of the block.")
        for idx, block_op in enumerate(self.ops):
            if block_op is op:
                return idx
        assert False, "Unexpected xdsl error"

    def detach_op(self, op: int | Operation) -> Operation:
        """
        Detach an operation from the block.
        Returns the detached operation.
        """
        if isinstance(op, Operation):
            op_idx = self.get_operation_index(op)
        else:
            op_idx = op
            op = self.ops[op_idx]
        if op.parent is not self:
            raise Exception("Cannot detach operation from a different block.")
        op.parent = None
        self.ops = self.ops[:op_idx] + self.ops[op_idx + 1:]
        return op

    def erase_op(self, op: int | Operation, safe_erase: bool = True) -> None:
        """
        Erase an operation from the block.
        If safe_erase is True, check that the operation has no uses.
        """
        op = self.detach_op(op)
        op.erase(safe_erase=safe_erase)

    def walk(self, fun: Callable[[Operation], None]) -> None:
        """Call a function on all operations contained in the block."""
        for op in self.ops:
            op.walk(fun)

    def verify(self) -> None:
        for operation in self.ops:
            if operation.parent != self:
                raise Exception(
                    "Parent pointer of operation does not refer to containing region"
                )
            operation.verify()

    def drop_all_references(self) -> None:
        """
        Drop all references to other operations.
        This function is called prior to deleting a block.
        """
        self.parent = None
        for op in self.ops:
            op.drop_all_references()

    def erase(self, safe_erase: bool = True) -> None:
        """
        Erase the block, and remove all its references to other operations.
        If safe_erase is specified, check that no operation results are used outside
        the block.
        """
        assert self.parent is None, "Blocks with parents should first be detached " + \
                                    "before erasure."
        self.drop_all_references()
        for op in self.ops:
            op.erase(safe_erase=safe_erase, drop_references=False)

    def is_structurally_equivalent(
            self,
            other: IRNode,
            context: Optional[dict[IRNode, IRNode]] = None) -> bool:
        """
        Check if two blocks are structurally equivalent.
        The context is a mapping of IR nodes to IR nodes that are already known to be equivalent.
        This enables checking whether the use dependencies and successors are equivalent.
        """
        if context is None:
            context = {}
        if not isinstance(other, Block):
            return False
        if len(self.args) != len(other.args) or \
            len(self.ops) != len(other.ops):
            return False
        for arg, other_arg in zip(self.args, other.args):
            if arg.typ != other_arg.typ:
                return False
            context[arg] = other_arg
        # Add self to the context so Operations can check for identical parents
        context[self] = other
        if not all(
                op.is_structurally_equivalent(other_op, context)
                for op, other_op in zip(self.ops, other.ops)):
            return False

        return True

    def __eq__(self, other: object) -> bool:
        return self is other

    def __hash__(self) -> int:
        return id(self)


@dataclass
class Region(IRNode):
    """A region contains a CFG of blocks. Regions are contained in operations."""

    blocks: list[Block] = field(default_factory=list, init=False)
    """Blocks contained in the region. The first block is the entry block."""

    parent: Operation | None = field(default=None, init=False, repr=False)
    """Operation containing the region."""

    def parent_block(self) -> Block | None:
        return self.parent.parent if self.parent else None

    def parent_op(self) -> Operation | None:
        return self.parent

    def parent_region(self) -> Region | None:
        return self.parent.parent.parent if self.parent and self.parent.parent else None

    def __repr__(self) -> str:
        return f"Region(num_blocks={len(self.blocks)})"

    @staticmethod
    def from_operation_list(ops: list[Operation]) -> Region:
        block = Block.from_ops(ops)
        region = Region()
        region.add_block(block)
        return region

    @staticmethod
    def from_block_list(blocks: list[Block]) -> Region:
        region = Region()
        for block in blocks:
            region.add_block(block)
        return region

    @staticmethod
    def get(arg: Region | list[Block] | list[Operation]) -> Region:
        if isinstance(arg, Region):
            return arg
        if isinstance(arg, list):
            if len(arg) == 0:
                return Region.from_operation_list([])
            if isinstance(arg[0], Block):
                return Region.from_block_list(cast(list[Block], arg))
            if isinstance(arg[0], Operation):
                return Region.from_operation_list(cast(list[Operation], arg))
        raise TypeError(f"Can't build a region with argument {arg}")

    @property
    def ops(self) -> list[Operation]:
        """
        Get the operations of a single-block region.
        Returns an exception if the region is not single-block.
        """
        if len(self.blocks) != 1:
            raise ValueError(
                "'ops' property of Region class is only available "
                "for single-block regions.")
        return self.blocks[0].ops

    @property
    def op(self) -> Operation:
        """
        Get the operation of a single-operation single-block region.
        Returns an exception if the region is not single-operation single-block.
        """
        if len(self.blocks) != 1 or len(self.blocks[0].ops) != 1:
            raise ValueError("'op' property of Region class is only available "
                             "for single-operation single-block regions.")
        return self.blocks[0].ops[0]

    def _attach_block(self, block: Block) -> None:
        """Attach a block to the region, and check that it has no parents."""
        if block.parent:
            raise ValueError(
                "Can't add to a region a block already attached to a region.")
        if block.is_ancestor(self):
            raise ValueError(
                "Can't add a block to a region contained in the block.")
        block.parent = self

    def add_block(self, block: Block) -> None:
        """Add a block to the region."""
        self._attach_block(block)
        self.blocks.append(block)

    def insert_block(self, blocks: Block | list[Block], index: int) -> None:
        """
        Insert one or multiple blocks at a given index in the region.
        The blocks should not be attached to another region.
        """
        if index < 0 or index > len(self.blocks):
            raise ValueError(
                f"Can't insert block in index {index} in a block with "
                f"{len(self.blocks)} blocks.")
        if not isinstance(blocks, list):
            blocks = [blocks]
        for block in blocks:
            self._attach_block(block)
        self.blocks = self.blocks[:index] + blocks + self.blocks[index:]

    def get_block_index(self, block: Block) -> int:
        """Get the block position in a region."""
        if block.parent is not self:
            raise Exception("Block is not a child of the region.")
        for idx, region_block in enumerate(self.blocks):
            if region_block is block:
                return idx
        assert False, "Unexpected xdsl error"

    def detach_block(self, block: int | Block) -> Block:
        """
        Detach a block from the region.
        Returns the detached block.
        """
        if isinstance(block, Block):
            block_idx = self.get_block_index(block)
        else:
            block_idx = block
            block = self.blocks[block_idx]
        block.parent = None
        self.blocks = self.blocks[:block_idx] + self.blocks[block_idx + 1:]
        return block

    def erase_block(self, block: int | Block, safe_erase: bool = True) -> None:
        """
        Erase a block from the region.
        If safe_erase is True, check that the block has no uses.
        """
        block = self.detach_block(block)
        block.erase(safe_erase=safe_erase)

    def clone_into(self,
                   dest: Region,
                   insert_index: int | None = None,
                   value_mapper: dict[SSAValue, SSAValue] | None = None,
                   block_mapper: dict[Block, Block] | None = None):
        """
        Clone all block of this region into `dest` to position `insert_index`
        """
        assert dest and dest != self
        if insert_index is None:
            insert_index = len(dest.blocks)
        if value_mapper is None:
            value_mapper = {}
        if block_mapper is None:
            block_mapper = {}

        for block in self.blocks:
            new_block = Block()
            block_mapper[block] = new_block
            for idx, block_arg in enumerate(block.args):
                new_block.insert_arg(block_arg.typ, idx)
                value_mapper[block_arg] = new_block.args[idx]
            for op in block.ops:
                new_block.add_op(op.clone(value_mapper, block_mapper))
            dest.insert_block(new_block, insert_index)
            insert_index += 1

    def walk(self, fun: Callable[[Operation], None]) -> None:
        """Call a function on all operations contained in the region."""
        for block in self.blocks:
            block.walk(fun)

    def verify(self) -> None:
        for block in self.blocks:
            block.verify()
            if block.parent != self:
                raise Exception(
                    "Parent pointer of block does not refer to containing region"
                )

    def drop_all_references(self) -> None:
        """
        Drop all references to other operations.
        This function is called prior to deleting a region.
        """
        self.parent = None
        for block in self.blocks:
            block.drop_all_references()

    def erase(self) -> None:
        """
        Erase the region, and remove all its references to other operations.
        """
        assert self.parent, "Regions with parents should first be " + \
                            "detached before erasure."
        self.drop_all_references()

    def move_blocks(self, region: Region) -> None:
        """
        Move the blocks of this region to another region. Leave no blocks in this region.
        """
        region.blocks = self.blocks
        self.blocks = []
        for block in region.blocks:
            block.parent = region

    def is_structurally_equivalent(
            self,
            other: IRNode,
            context: Optional[dict[IRNode, IRNode]] = None) -> bool:
        """
        Check if two regions are structurally equivalent.
        The context is a mapping of IR nodes to IR nodes that are already known to be equivalent.
        This enables checking whether the use dependencies and successors are equivalent.
        """
        if context is None:
            context = {}
        if not isinstance(other, Region):
            return False
        if len(self.blocks) != len(other.blocks):
            return False
        # register all blocks in the context so we can check whether ops have
        # the corrects successors
        for block, other_block in zip(self.blocks, other.blocks):
            context[block] = other_block
        if not all(
                block.is_structurally_equivalent(other_block, context)
                for block, other_block in zip(self.blocks, other.blocks)):
            return False
        return True
